from tests.utils import create_and_write_both_type_of_functions
from tests.materials import Nickel, Argon


def create_nickel():
    ni = Nickel()
    create_and_write_both_type_of_functions(ni.lammps_gr_filename,
                                            ni.real_space_filename,
                                            ni.reciprocal_space_filename,
                                            **ni.kwargs)


def create_argon():
    ar = Argon()
    create_and_write_both_type_of_functions(ar.lammps_gr_filename,
                                            ar.real_space_filename,
                                            ar.reciprocal_space_filename,
                                            **ar.kwargs)


if __name__ == "__main__":
    create_nickel()
    create_argon()



import numpy as np
import os
import tempfile
import unittest

from tests.utils import get_data_path, get_index_of_function, load_data
from tests.materials import Argon
from pystog.utils import RealSpaceHeaders, ReciprocalSpaceHeaders
from pystog.stog import NoInputFilesException, StoG


class TestStogBase(unittest.TestCase):
    rtol = 1.0
    atol = 1.0

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the tolerance
        self.first = self.material.reciprocal_space_first
        self.last = self.material.reciprocal_space_last

        data = load_data(self.material.reciprocal_space_filename)
        self.q = data[:, get_index_of_function("Q", ReciprocalSpaceHeaders)]
        self.sq = data[:, get_index_of_function("S(Q)", ReciprocalSpaceHeaders)]
        self.fq = data[
            :, get_index_of_function("Q[S(Q)-1]", ReciprocalSpaceHeaders)
        ]
        self.fq_keen = data[
            :, get_index_of_function("FK(Q)", ReciprocalSpaceHeaders)
        ]
        self.dcs = data[
            :, get_index_of_function("DCS(Q)", ReciprocalSpaceHeaders)
        ]

        # targets for 1st peaks
        self.sq_target = self.material.sq_target
        self.dsq_target = 3.087955
        self.fq_target = self.material.fq_target
        self.fq_keen_target = self.material.fq_keen_target
        self.dcs_target = self.material.dcs_target

        # setup the first, last indices
        self.real_space_first = self.material.real_space_first
        self.real_space_last = self.material.real_space_last

        data = load_data(self.material.real_space_filename)
        self.r = data[:, get_index_of_function("r", RealSpaceHeaders)]
        self.gofr = data[:, get_index_of_function("g(r)", RealSpaceHeaders)]
        self.GofR = data[:, get_index_of_function("G(r)", RealSpaceHeaders)]
        self.GKofR = data[:, get_index_of_function("GK(r)", RealSpaceHeaders)]

        self.gofr_target = self.material.gofr_target
        self.GofR_target = self.material.GofR_target
        self.GKofR_target = self.material.GKofR_target

        self.gofr_ff_target = self.material.gofr_ff_target
        self.GofR_ff_target = self.material.GofR_ff_target
        self.GKofR_ff_target = self.material.GKofR_ff_target

        self.gofr_lorch_target = self.material.gofr_lorch_target
        self.GofR_lorch_target = self.material.GofR_lorch_target
        self.GKofR_lorch_target = self.material.GKofR_lorch_target

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.material = Argon()
        self.initialize_material()

        self.real_xtarget = 3.525
        self.reciprocal_xtarget = 1.94
        self.fourier_filter_cutoff = 1.5

        filename = get_data_path(self.material.reciprocal_space_filename)
        self.kwargs_for_files = {
            "Files": [
                {
                    "Filename": filename,
                    "ReciprocalFunction": "S(Q)",
                    "Qmin": 0.02,
                    "Qmax": 15.0,
                    "Y": {"Offset": 0.0, "Scale": 1.0},
                    "X": {"Offset": 0.0},
                },
                {
                    "Filename": filename,
                    "ReciprocalFunction": "S(Q)",
                    "Qmin": 1.90,
                    "Qmax": 35.2,
                    "Y": {"Offset": 0.0, "Scale": 1.0},
                    "X": {"Offset": 0.0},
                },
            ]
        }

        self.kwargs_for_stog_input = {
            "NumberDensity": self.material.kwargs["rho"],
            "<b_coh>^2": self.material.kwargs["<b_coh>^2"],
            "<b_tot^2>": self.material.kwargs["<b_tot^2>"],
            "FourierFilter": {"Cutoff": self.fourier_filter_cutoff},
            "OmittedXrangeCorrection": False,
            "Rdelta": self.r[1] - self.r[0],
            "Rmin": min(self.r),
            "Rmax": max(self.r),
        }

    def tearDown(self):
        unittest.TestCase.tearDown(self)


class TestStogInit(TestStogBase):
    def setUp(self):
        super(TestStogInit, self).setUp()

    def test_stog_init(self):
        stog = StoG()
        self.assertEqual(stog.xmin, 100)
        self.assertEqual(stog.xmax, 0)
        self.assertEqual(stog.qmin, None)
        self.assertEqual(stog.qmax, None)
        self.assertEqual(stog.files, None)
        self.assertEqual(stog.sq_title, "S(Q) Merged")
        self.assertEqual(stog.qsq_minus_one_title, "Q[S(Q)-1] Merged")
        self.assertEqual(stog.sq_ft_title, "S(Q) FT")
        self.assertEqual(stog.fq_title, "F(Q) Merged")
        self.assertEqual(stog.real_space_function, "g(r)")
        self.assertEqual(stog.gr_title, "g(r) Merged")
        self.assertEqual(stog.gr_ft_title, "g(r) FT")
        self.assertEqual(stog.gr_lorch_title, "g(r) FT Lorched")
        self.assertEqual(stog.GKofR_title, "G(r) (Keen Version)")
        self.assertEqual(stog.rmin, 0.0)
        self.assertEqual(stog.rmax, 50.0)
        self.assertEqual(stog.rdelta, 0.01)
        self.assertEqual(stog.density, 1.0)
        self.assertEqual(stog.bcoh_sqrd, 1.0)
        self.assertEqual(stog.btot_sqrd, 1.0)
        self.assertEqual(stog.low_q_correction, False)
        self.assertEqual(stog.lorch_flag, False)
        self.assertEqual(stog.fourier_filter_cutoff, None)
        self.assertEqual(stog.merged_opts, {"Y": {"Offset": 0.0, "Scale": 1.0}})
        self.assertEqual(stog.stem_name, "out")
        self.assertEqual(stog.reciprocal_individuals.size, 0)
        self.assertEqual(stog.q_master, {})
        self.assertEqual(stog.sq_master, {})
        self.assertEqual(stog.sq_individuals.size, 0)
        self.assertEqual(stog.r_master, {})
        self.assertEqual(stog.gr_master, {})

    def test_stog_init_kwargs_files(self):
        stog = StoG(**{"Files": ["file1.txt", "file2.txt"]})
        self.assertEqual(stog.files, ["file1.txt", "file2.txt"])

    def test_stog_init_kwargs_real_space_function(self):
        stog = StoG(**{"RealSpaceFunction": "G(r)"})
        self.assertEqual(stog.real_space_function, "G(r)")

    def test_stog_init_kwargs_rmin(self):
        stog = StoG(**{"Rmin": 2.0})
        self.assertEqual(stog.rmin, 2.0)

    def test_stog_init_kwargs_rmax(self):
        stog = StoG(**{"Rmax": 25.0})
        self.assertEqual(stog.rmax, 25.0)

    def test_stog_init_kwargs_rdelta(self):
        stog = StoG(**{"Rdelta": 0.5})
        self.assertEqual(stog.rdelta, 0.5)

    def test_stog_init_kwargs_rpoints(self):
        stog = StoG(**{"Rpoints": 250})
        self.assertEqual(stog.rdelta, 0.2)

    def test_stog_init_kwargs_density(self):
        stog = StoG(**{"NumberDensity": 2.0})
        self.assertEqual(stog.density, 2.0)

    def test_stog_init_kwargs_low_q_correction(self):
        stog = StoG(**{"OmittedXrangeCorrection": True})
        self.assertEqual(stog.low_q_correction, True)

    def test_stog_init_kwargs_lorch_flag(self):
        stog = StoG(**{"LorchFlag": True})
        self.assertEqual(stog.lorch_flag, True)

    def test_stog_init_kwargs_fourier_filter_cutoff(self):
        stog = StoG(**{"FourierFilter": {"Cutoff": 1.0}})
        self.assertEqual(stog.fourier_filter_cutoff, 1.0)

    def test_stog_init_kwargs_bcoh_sqrd(self):
        stog = StoG(**{"<b_coh>^2": 0.5})
        self.assertEqual(stog.bcoh_sqrd, 0.5)

    def test_stog_init_kwargs_btot_sqrd(self):
        stog = StoG(**{"<b_tot^2>": 0.75})
        self.assertEqual(stog.btot_sqrd, 0.75)

    def test_stog_init_kwargs_qmin_only(self):
        stog = StoG(**{"Merging": {"Transform": {"Qmin": 0.5}}})
        self.assertEqual(stog.qmin, 0.5)
        self.assertEqual(stog.qmax, None)

    def test_stog_init_kwargs_qmax_only(self):
        stog = StoG(**{"Merging": {"Transform": {"Qmax": 30.0}}})
        self.assertEqual(stog.qmin, None)
        self.assertEqual(stog.qmax, 30.0)

    def test_stog_init_kwargs_qmin_and_qmax(self):
        stog = StoG(**{"Merging": {"Transform": {"Qmin": 0.5, "Qmax": 30.0}}})
        self.assertEqual(stog.qmin, 0.5)
        self.assertEqual(stog.qmax, 30.0)

    def test_stog_init_kwargs_output_stem_name(self):
        stog = StoG(**{"Outputs": {"StemName": "myName"}})
        self.assertEqual(stog.stem_name, "myName")


class TestStogAttributes(TestStogBase):
    def setUp(self):
        super(TestStogAttributes, self).setUp()

    def test_stog_xmin_setter(self):
        stog = StoG()
        stog.xmin = 0.25
        self.assertEqual(stog.xmin, 0.25)

    def test_stog_xmax_setter(self):
        stog = StoG()
        stog.xmax = 10.0
        self.assertEqual(stog.xmax, 10.0)

    def test_stog_dr_getter(self):
        stog = StoG()
        self.assertAlmostEqual(stog.dr[0], 0.0)
        self.assertAlmostEqual(stog.dr[-1], 50.0)

    def test_stog_dr_setter_rmax(self):
        stog = StoG()
        stog.rmax = 25.0
        self.assertAlmostEqual(stog.dr[0], 0.0)
        self.assertAlmostEqual(stog.dr[-1], 25.0)

    def test_stog_dr_setter_rmin(self):
        stog = StoG()
        stog.rmin = 10.0
        self.assertAlmostEqual(stog.dr[0], 10.0)
        self.assertAlmostEqual(stog.dr[-1], 50.0)

    def test_stog_dr_setter_rdelta(self):
        stog = StoG()
        stog.rdelta = 0.5
        self.assertEqual(stog.dr[1] - stog.dr[0], 0.5)

    def test_stog_gr_title_function_setter(self):
        stog = StoG()
        stog.gr_title = "G(r) dog"
        self.assertEqual(stog.gr_title, "G(r) dog")

    def test_stog_gr_ft_title_function_setter(self):
        stog = StoG()
        stog.gr_ft_title = "G(r) FT dog"
        self.assertEqual(stog.gr_ft_title, "G(r) FT dog")

    def test_stog_gr_lorch_title_function_setter(self):
        stog = StoG()
        stog.gr_lorch_title = "G(r) FT Lorch dog"
        self.assertEqual(stog.gr_lorch_title, "G(r) FT Lorch dog")

    def test_stog_GKofR_title_function_setter(self):
        stog = StoG()
        stog.GKofR_title = "GK(r) dog"
        self.assertEqual(stog.GKofR_title, "GK(r) dog")

    def test_stog_sq_title_function_setter(self):
        stog = StoG()
        stog.sq_title = "S(Q) dog"
        self.assertEqual(stog.sq_title, "S(Q) dog")

    def test_stog_fq_title_function_setter(self):
        stog = StoG()
        stog.fq_title = "F(Q) dog"
        self.assertEqual(stog.fq_title, "F(Q) dog")

    def test_stog_qsq_minus_one_title_setter(self):
        stog = StoG()
        stog.qsq_minus_one_title = "Q[S(Q)-1] dog"
        self.assertEqual(stog.qsq_minus_one_title, "Q[S(Q)-1] dog")

    def test_stog_sq_ft_title_setter(self):
        stog = StoG()
        stog.sq_ft_title = "S(Q) FT dog"
        self.assertEqual(stog.sq_ft_title, "S(Q) FT dog")

    def test_stog_real_space_function_setter(self):
        stog = StoG()
        stog.real_space_function = "GK(r)"
        self.assertEqual(stog.real_space_function, "GK(r)")
        self.assertEqual(stog.gr_title, "GK(r) Merged")
        self.assertEqual(stog.real_space_function, "GK(r)")
        self.assertEqual(stog.gr_ft_title, "GK(r) FT")
        self.assertEqual(stog.gr_lorch_title, "GK(r) FT Lorched")

    def test_stog_low_q_correction_exception(self):
        stog = StoG()
        with self.assertRaises(TypeError):
            stog.low_q_correction = 1.0

    def test_stog_lorch_flag_exception(self):
        stog = StoG()
        with self.assertRaises(TypeError):
            stog.lorch_flag = 1.0

    def test_stog_real_space_function_exception(self):
        stog = StoG()
        with self.assertRaises(ValueError):
            stog.real_space_function = "Dog"


class TestStogStorageArrays(TestStogBase):
    def setUp(self):
        super(TestStogStorageArrays, self).setUp()
        self.target = np.random.randn(3, 10)

    def test_stog_reciprocal_individuals_setter(self):
        stog = StoG()
        stog.reciprocal_individuals = self.target
        np.testing.assert_allclose(stog.reciprocal_individuals, self.target)

    def test_stog_sq_individuals_setter(self):
        stog = StoG()
        stog.sq_individuals = self.target
        np.testing.assert_allclose(stog.sq_individuals, self.target)

    def test_stog_q_master_setter(self):
        stog = StoG()
        stog.q_master = self.target
        np.testing.assert_allclose(stog.q_master, self.target)

    def test_stog_sq_master_setter(self):
        stog = StoG()
        stog.sq_master = self.target
        np.testing.assert_allclose(stog.sq_master, self.target)

    def test_stog_r_master_setter(self):
        stog = StoG()
        stog.r_master = self.target
        np.testing.assert_allclose(stog.r_master, self.target)

    def test_stog_gr_master_setter(self):
        stog = StoG()
        stog.gr_master = self.target
        np.testing.assert_allclose(stog.gr_master, self.target)


class TestStogGeneralMethods(TestStogBase):
    def setUp(self):
        super(TestStogGeneralMethods, self).setUp()

    def test_stog_append_file(self):
        stog = StoG(**{"Files": ["file1.txt", "file2.txt"]})
        stog.append_file("file3.txt")
        self.assertEqual(stog.files, ["file1.txt", "file2.txt", "file3.txt"])

    def test_stog_extend_file_list(self):
        stog = StoG(**{"Files": ["file1.txt", "file2.txt"]})
        stog.extend_file_list(["file3.txt", "file4.txt"])
        self.assertEqual(
            stog.files, ["file1.txt", "file2.txt", "file3.txt", "file4.txt"]
        )


class TestStogDatasetSpecificMethods(TestStogBase):
    def setUp(self):
        super(TestStogDatasetSpecificMethods, self).setUp()

    def test_stog_apply_scales_and_offset(self):
        q, sq, dq = StoG.apply_scales_and_offset(self.q, self.sq)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, np.zeros_like(sq))

    def test_stog_apply_scales_and_offset_with_dy(self):
        q, sq, dq = StoG.apply_scales_and_offset(self.q, self.sq, dy=self.sq)
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_apply_scales_and_offset_with_yscale(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, yscale=2.0
        )
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq * 2.0)
        np.testing.assert_allclose(dq, self.sq * 2.0)

    def test_stog_apply_scales_and_offset_with_yoffset(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, yoffset=2.0
        )
        np.testing.assert_allclose(q, self.q)
        np.testing.assert_allclose(sq, self.sq + 2.0)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_apply_scales_and_offset_with_xoffset(self):
        q, sq, dq = StoG.apply_scales_and_offset(
            self.q, self.sq, dy=self.sq, xoffset=2.0
        )
        np.testing.assert_allclose(q, self.q + 2.0)
        np.testing.assert_allclose(sq, self.sq)
        np.testing.assert_allclose(dq, self.sq)

    def test_stog_add_dataset(self):
        # Number of decimal places for precision
        places = 5

        # Initialize with material info for Argon
        stog = StoG(
            **{
                "<b_coh>^2": self.kwargs["<b_coh>^2"],
                "<b_tot^2>": self.kwargs["<b_tot^2>"],
            }
        )

        # Add the S(Q) data set and check values against targets
        info = {"data": [self.q, self.sq], "ReciprocalFunction": "S(Q)"}
        stog.add_dataset(info)
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget
        )
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places,
        )
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the Q[S(Q)-1] data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {"data": [self.q, self.fq], "ReciprocalFunction": "Q[S(Q)-1]"}
        stog.add_dataset(info)

        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.fq_target[0],
            places=places,
        )
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride], 0.0
        )
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the FK(Q) data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {"data": [self.q, self.fq_keen], "ReciprocalFunction": "FK(Q)"}
        stog.add_dataset(info)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.fq_keen_target[0],
            places=places,
        )
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride], 0.0
        )
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

        # Add the DCS(Q) data set and check values for it and S(Q) against
        # targets
        stride = stog.reciprocal_individuals.shape[1]
        info = {"data": [self.q, self.dcs], "ReciprocalFunction": "DCS(Q)"}
        stog.add_dataset(info)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first + stride],
            self.dcs_target[0],
            places=places,
        )
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first + stride], 0.0
        )
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yscale(self):
        # Scale S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            "data": [self.q, self.sq],
            "ReciprocalFunction": "S(Q)",
            "Y": {"Scale": 2.0},
        }
        stog.add_dataset(info)
        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first], self.sq_target[0]
        )
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertNotEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0]
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yoffset(self):
        # Offset S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            "data": [self.q, self.sq],
            "ReciprocalFunction": "S(Q)",
            "Y": {"Offset": 2.0},
        }
        stog.add_dataset(info)
        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first], self.sq_target[0]
        )
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertNotEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0]
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_yscale_with_dy(self):
        # Scale S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            "data": [self.q, self.sq, self.sq],
            "ReciprocalFunction": "S(Q)",
            "Y": {"Scale": 2.0},
        }
        stog.add_dataset(info)

        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first], self.sq_target[0]
        )

        dsq_target = info["Y"]["Scale"] * 2.59173
        self.assertEqual(stog.reciprocal_individuals[2][self.first], dsq_target)

        self.assertNotEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0]
        )

        self.assertEqual(stog.sq_individuals[2][self.first], dsq_target)

    def test_stog_add_dataset_yoffset_with_dy(self):
        # Offset S(Q) and make sure it does not equal original target values
        stog = StoG()
        info = {
            "data": [self.q, self.sq, self.sq],
            "ReciprocalFunction": "S(Q)",
            "Y": {"Offset": 2.0},
        }
        stog.add_dataset(info)

        self.assertNotEqual(
            stog.reciprocal_individuals[1][self.first], self.sq_target[0]
        )

        dsq_target = 2.59173
        self.assertEqual(stog.reciprocal_individuals[2][self.first], dsq_target)

        self.assertNotEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0]
        )

        self.assertEqual(stog.sq_individuals[2][self.first], dsq_target)

    def test_stog_add_dataset_xoffset(self):
        # Offset Q from 1.96 -> 2.14
        stog = StoG()
        info = {
            "data": [self.q, self.sq],
            "ReciprocalFunction": "S(Q)",
            "X": {"Offset": 0.2},
        }
        stog.add_dataset(info)
        self.assertEqual(stog.reciprocal_individuals[0][self.first], 2.14)

    def test_stog_add_dataset_qmin_qmax_crop(self):
        # Check qmin and qmax apply cropping
        stog = StoG()
        info = {"data": [self.q, self.sq], "ReciprocalFunction": "S(Q)"}
        stog.qmin = 1.5
        stog.qmax = 12.0
        stog.add_dataset(info)
        self.assertEqual(stog.reciprocal_individuals[0][0], stog.qmin)
        self.assertEqual(stog.reciprocal_individuals[0][-1], stog.qmax)

    def test_stog_add_dataset_default_reciprocal_space_function(self):
        # Checks the default reciprocal space function is S(Q)
        places = 5
        stog = StoG()
        info = {"data": [self.q, self.sq]}
        stog.add_dataset(info)
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget
        )
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places,
        )
        self.assertEqual(stog.sq_individuals[2][self.first], 0.0)

    def test_stog_add_dataset_wrong_reciprocal_space_function_exception(self):
        # Check qmin and qmax apply cropping
        stog = StoG()
        info = {"data": [self.q, self.sq], "ReciprocalFunction": "ABCDEFG(Q)"}
        with self.assertRaises(ValueError):
            stog.add_dataset(info)

    def test_stog_read_nexus_file_by_bank(self):
        stog = StoG()
        bank = 1
        nexus_file = "./tests/test_data/nexus/pystog_test.nxs"

        stog.read_nexus_file_by_bank(nexus_file, bank, "pystog_test")

        self.compareResults(
            "./outpystog_test_bank1.dat",
            "./tests/test_data/nexus/pystog_test_bank1.dat",
        )

    def compareResults(self, current, expected, cleanup=True):
        output = np.loadtxt(current, unpack=True, skiprows=2)
        expected_output = np.loadtxt(expected, unpack=True, skiprows=2)

        np.allclose(output, expected_output)
        if cleanup and os.path.exists(current):
            os.remove(current)

    def test_stog_read_all_nexus_file_by_json(self):
        self.nexus_kwargs = {
            "NexusFile": "./tests/test_data/nexus/pystog_test.nxs",
            "WorkspaceName": "pystog_test",
            "Files": [
                {"BankNumber": "2"},
                {"BankNumber": "3"},
                {"BankNumber": "4"},
                {"BankNumber": "5"},
            ],
            "Outputs": {"StemName": "stem_"},
        }
        stog = StoG(**self.nexus_kwargs)
        stog.read_all_nexus_file_banks()
        for i in (2, 3, 4, 5):
            self.compareResults(
                "./stem_pystog_test_bank{}.dat".format(i),
                "./tests/test_data/nexus/pystog_test_bank{}.dat".format(i),
            )

    def test_stog_read_dataset(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(
            **{
                "<b_coh>^2": self.kwargs["<b_coh>^2"],
                "<b_tot^2>": self.kwargs["<b_tot^2>"],
            }
        )
        info = {
            "Filename": get_data_path(self.material.reciprocal_space_filename),
            "ReciprocalFunction": "S(Q)",
            "Qmin": 0.02,
            "Qmax": 35.2,
            "Y": {"Offset": 0.0, "Scale": 1.0},
            "X": {"Offset": 0.0},
        }

        info["index"] = 0
        stog.read_dataset(info)

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget
        )
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places,
        )
        self.assertEqual(
            stog.reciprocal_individuals[2][self.first], self.dsq_target
        )
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )
        self.assertEqual(stog.sq_individuals[2][self.first], self.dsq_target)

    def test_stog_read_dataset_xcol_data_format_exception(self):
        stog = StoG()
        filename = get_data_path(self.material.reciprocal_space_filename)
        with self.assertRaises(RuntimeError):
            stog.read_dataset({"Filename": filename}, xcol=99)

    def test_stog_read_dataset_ycol_data_format_exception(self):
        stog = StoG()
        filename = get_data_path(self.material.reciprocal_space_filename)
        with self.assertRaises(RuntimeError):
            stog.read_dataset({"Filename": filename}, ycol=99)

    def test_stog_read_dataset_dycol_too_large(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(
            **{
                "<b_coh>^2": self.kwargs["<b_coh>^2"],
                "<b_tot^2>": self.kwargs["<b_tot^2>"],
            }
        )
        info = {
            "Filename": get_data_path(self.material.reciprocal_space_filename),
            "ReciprocalFunction": "S(Q)",
            "Qmin": 0.02,
            "Qmax": 35.2,
            "Y": {"Offset": 0.0, "Scale": 1.0},
            "X": {"Offset": 0.0},
        }

        info["index"] = 0
        stog.read_dataset(info, dycol=99)

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget
        )
        self.assertAlmostEqual(
            stog.reciprocal_individuals[1][self.first],
            self.sq_target[0],
            places=places,
        )
        self.assertEqual(stog.reciprocal_individuals[2][self.first], 0.0)
        self.assertAlmostEqual(
            stog.sq_individuals[1][self.first], self.sq_target[0], places=places
        )

    def test_stog_read_all_data_assertion(self):
        stog = StoG()
        with self.assertRaises(NoInputFilesException):
            stog.read_all_data()

        stog.files = list()
        with self.assertRaises(NoInputFilesException):
            stog.read_all_data()

    def test_stog_read_all_data_for_files_length(self):
        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()

        # Check S(Q) data against targets
        self.assertEqual(len(stog.files), len(self.kwargs_for_files["Files"]))

    def test_stog_read_all_data(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()

        # Check S(Q) data against targets
        self.assertEqual(
            stog.reciprocal_individuals[0][self.first], self.reciprocal_xtarget
        )
        for index in range(len(stog.files)):
            self.assertAlmostEqual(
                stog.reciprocal_individuals[1][self.first],
                self.sq_target[0],
                places=places,
            )
            self.assertAlmostEqual(
                stog.sq_individuals[1][self.first],
                self.sq_target[0],
                places=places,
            )

    def test_stog_merge_data(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()

        # Check S(Q) data against targets
        self.assertEqual(
            stog.q_master[stog.sq_title][self.first], self.reciprocal_xtarget
        )
        self.assertAlmostEqual(
            stog.sq_master[stog.sq_title][self.first],
            self.sq_target[0],
            places=places,
        )

    def test_stog_merge_data_qsq_opts_scale(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        qsq_opts = {"Y": {"Scale": 2.0}}

        # Test Q[S(Q)-1] scale
        stog.merged_opts["Q[S(Q)-1]"] = qsq_opts
        stog.merge_data()
        self.assertAlmostEqual(
            stog.sq_master[stog.qsq_minus_one_title][self.first],
            qsq_opts["Y"]["Scale"] * self.fq_target[0],
            places=places,
        )

    def test_stog_merge_data_qsq_opts_offset(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG()
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        qsq_opts = {"Y": {"Offset": 1.0}}

        # Test Q[S(Q)-1] scale
        stog.merged_opts["Q[S(Q)-1]"] = qsq_opts
        stog.merge_data()
        self.assertAlmostEqual(
            stog.sq_master[stog.qsq_minus_one_title][self.first],
            qsq_opts["Y"]["Offset"] + self.fq_target[0],
            places=places,
        )


class TestStogTransformSpecificMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogTransformSpecificMethods, self).setUp()
        self.lowR_target = 0.4720653

    def test_stog_transform_merged_default(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.gofr_target[0],
            places=places,
        )

        # Test if no dr defined
        stog.dr = None
        stog.transform_merged()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.gofr_target[0],
            places=places,
        )

    def test_stog_transform_merged_GofR(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check G(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.GofR_target[0],
            places=places,
        )

    def test_stog_transform_merged_GKofR(self):
        # Number of decimal places for precision
        places = 2

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        # Check GK(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_title][self.real_space_first],
            self.GKofR_target[0],
            places=places,
        )

    def test_stog_transform_merged_for_nan_after_filter(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.assertFalse(np.isnan(stog.sq_master[stog.sq_title]).any())
        self.assertFalse(np.isnan(stog.gr_master[stog.gr_title]).any())

    def test_stog_fourier_filter(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.gofr_ff_target[0],
            places=places,
        )

    def test_stog_fourier_filter_before_transform_merged_call(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.gofr_ff_target[0],
            places=places,
        )

    def test_stog_fourier_filter_GofR(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.GofR_ff_target[0],
            places=places,
        )

    def test_stog_fourier_filter_GKofR(self):
        # Number of decimal places for precision
        places = 1

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        # Check g(r) data against targets
        self.assertAlmostEqual(
            stog.r_master[stog.gr_ft_title][self.real_space_first],
            self.real_xtarget,
            places=places,
        )
        self.assertAlmostEqual(
            stog.gr_master[stog.gr_ft_title][self.real_space_first],
            self.GKofR_ff_target[0],
            places=places,
        )

    def test_stog_fourier_filter_for_nan_after_filter(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        stog.fourier_filter()

        self.assertFalse(np.isnan(stog.gr_master[stog.gr_title]).any())
        self.assertFalse(np.isnan(stog.sq_master[stog.sq_title]).any())
        self.assertFalse(np.isnan(stog.sq_master[stog._ft_title]).any())
        self.assertFalse(np.isnan(stog.sq_master[stog.sq_ft_title]).any())

    def test_stog_apply_lorch_default(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.gofr_lorch_target[0],
            places=places,
        )

    def test_stog_apply_lorch_GofR(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "G(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.GofR_lorch_target[0],
            places=places,
        )

    def test_stog_apply_lorch_GKofR(self):
        # Number of decimal places for precision
        places = 5

        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.real_space_function = "GK(r)"
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        stog.apply_lorch(q, sq, r)

        self.assertAlmostEqual(
            stog.gr_master[stog.gr_lorch_title][self.real_space_first],
            self.GKofR_lorch_target[0],
            places=places,
        )

    def test_stog_lowR_mean_square(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        gr = stog.gr_master[stog.gr_title]
        cost = stog._lowR_mean_square(stog.dr, gr)
        self.assertAlmostEqual(cost, self.lowR_target, places=7)

    def test_stog_get_lowR_mean_square(self):
        # Load S(Q) for Argon from test data
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        q, sq, r, gr = stog.fourier_filter()
        cost = stog._get_lowR_mean_square()
        self.assertAlmostEqual(cost, self.lowR_target, places=7)


class TestStogBookkeepingMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogBookkeepingMethods, self).setUp()
        stog = StoG(**self.kwargs_for_stog_input)
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.stog = stog

    def test_stog_add_keen_fq(self):
        stog = self.stog
        q = stog.sq_master[stog.sq_title]
        sq = stog.sq_master[stog.sq_title]
        stog._add_keen_fq(q, sq)
        self.assertTrue(stog.fq_title in stog.sq_master)

    def test_stog_add_keen_gr_default(self):
        stog = self.stog
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)

    def test_stog_add_keen_gr_GofR(self):
        stog = StoG(**self.kwargs_for_stog_input)
        stog.real_space_function = "G(r)"
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)

    def test_stog_add_keen_gr_GKofR(self):
        stog = StoG(**self.kwargs_for_stog_input)
        stog.real_space_function = "GK(r)"
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()
        r = stog.gr_master[stog.gr_title]
        gr = stog.gr_master[stog.gr_title]
        stog._add_keen_gr(r, gr)
        self.assertTrue(stog.GKofR_title in stog.gr_master)


class TestStogOutputMethods(TestStogDatasetSpecificMethods):
    def setUp(self):
        super(TestStogOutputMethods, self).setUp()
        stog = StoG(**self.kwargs_for_stog_input)
        stog.stem_name = "dog"
        stog.files = self.kwargs_for_files["Files"]
        stog.read_all_data()
        stog.merge_data()
        stog.transform_merged()

        self.stog = stog

    # Decorator to provide the data to run each write_out_<type> test
    def data_provider(self, x, y, filename):
        def write_out_decorator(write_out_func):
            def wrap_function(*args):
                # Using stem name
                write_out_func()

                data = {}
                data["x"], data["y"] = np.genfromtxt(
                    filename, skip_header=2, unpack=True
                )

                np.testing.assert_allclose(data["x"], x)
                np.testing.assert_allclose(
                    data["y"] - y,
                    np.zeros(len(y)),
                    rtol=2.0,
                    atol=2.0,
                    equal_nan=True,
                )

                os.remove(filename)

                # Using set filename
                tmp_filename = tempfile.mkstemp()[1]

                write_out_func(filename=tmp_filename)

                data["x"], data["y"] = np.genfromtxt(
                    tmp_filename, skip_header=2, unpack=True
                )

                np.testing.assert_allclose(data["x"], x)
                np.testing.assert_allclose(data["y"], y)
                os.remove(tmp_filename)

            return wrap_function

        return write_out_decorator

    # Tests
    def test_stog_write_df(self):
        outfile_path = tempfile.mkstemp()[1]
        x = self.stog.q_master[self.stog.sq_title]
        y = self.stog.sq_master[self.stog.sq_title]
        self.stog._write_out_to_file(x, y, outfile_path)
        data = {}
        data["x"], data["y"] = np.genfromtxt(
            outfile_path, skip_header=2, unpack=True
        )

        q = self.stog.q_master[self.stog.sq_title]
        sq = self.stog.sq_master[self.stog.sq_title]

        np.testing.assert_allclose(data["x"], q)
        np.testing.assert_allclose(data["y"], sq)
        os.remove(outfile_path)

    def test_write_out_merged_sq(self):
        # Have to decorate after the setUp() is called for the self.* args to
        # work
        @self.data_provider(
            self.stog.q_master[self.stog.sq_title],
            self.stog.sq_master[self.stog.sq_title],
            "dog.sq",
        )
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_merged_sq(*args, **kwargs)

        decorated_write_out_merged()

    def test_write_out_merged_gr(self):
        # Have to decorate after the setUp() is called for the self.* args to
        # work
        @self.data_provider(
            self.stog.r_master[self.stog.gr_title],
            self.stog.gr_master[self.stog.gr_title],
            "dog.gr",
        )
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_merged_gr(*args, **kwargs)

        decorated_write_out_merged()

    def test_write_out_ft_sq(self):
        self.stog.fourier_filter()
        # Have to decorate after the setUp() is called for the self.* args to
        # work

        @self.data_provider(
            self.stog.q_master[self.stog.sq_ft_title],
            self.stog.sq_master[self.stog.sq_ft_title],
            "dog_ft.sq",
        )
        def decorated_write_out_merged(*args, **kwargs):
            self.stog.write_out_ft_sq(*args, **kwargs)

        decorated_write_out_merged()


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



"""
========
pre_proc
========

This module is for some pre-processing of data
"""


class Pre_Proc:
    """
    The Pre_Proc class is used for performing some
    pre-processing task for the data

    :examples:

    >>> from pystog import Pre_Proc
    >>> pre_proc = Pre_Proc()
    >>> q, sq = numpy.loadtxt("my_sofq_file.txt", unpack=True)
    >>> q_rb, sq_rb = pre_proc.rebin(q, sq)
    """

    def __init__(self):
        pass

    def rebin(x, y, xmin, xdiv, xmax):
        """
        Rebin input data to the specified grid

        :param x: input X array
        :type x: numpy.array or list
        :param y: input Y array
        :type y: numpy.array or list
        :param xmin: Xmin
        :type xmin: float
        :param xdiv: X interval
        :type xdiv: float
        :param xmax: Xmax
        :type xmax: Xmax

        :return: (output X array, output Y array)
        :rtype: (list, list)
        """

        numpts = int((xmax - xmin) / xdiv) + 1
        xout = list()
        for loop in range(numpts):
            xout.append(xmin + loop * xdiv)
        yout = [0. for _ in range(len(xout) + 1)]
        ynorm = [0. for _ in range(len(xout) + 1)]

        for i, x_tmp in enumerate(x):
            if xmin <= x_tmp <= xmax:
                bin = int((x_tmp - xmin) / xdiv)
                scale1 = 1 - (x_tmp - xout[bin]) / xdiv
                scale2 = 1 - scale1

                yout[bin] += y[i] * scale1
                yout[bin + 1] += y[i] * scale2
                ynorm[bin] += scale1
                ynorm[bin + 1] += scale2

        for i in range(len(yout)):
            yout[i] /= ynorm[i]

        return (xout, yout[:-1])



import unittest
from numpy.testing import assert_allclose
from tests.utils import \
    load_data, get_index_of_function
from tests.materials import Nickel, Argon
from pystog.utils import \
    RealSpaceHeaders, ReciprocalSpaceHeaders
from pystog.fourier_filter import FourierFilter

# Real Space Function


class TestFourierFilterBase(unittest.TestCase):
    rtol = 0.2
    atol = 0.2

    def initialize_material(self):
        # setup input data
        self.kwargs = self.material.kwargs

        # setup the first, last indices
        self.real_space_first = self.material.real_space_first
        self.real_space_last = self.material.real_space_last

        data = load_data(self.material.real_space_filename)
        self.r = data[:, get_index_of_function("r", RealSpaceHeaders)]
        self.gofr = data[:, get_index_of_function("g(r)", RealSpaceHeaders)]
        self.GofR = data[:, get_index_of_function("G(r)", RealSpaceHeaders)]
        self.GKofR = data[:, get_index_of_function("GK(r)", RealSpaceHeaders)]

        # targets for 1st peaks
        self.gofr_ff_target = self.material.gofr_ff_target
        self.GofR_ff_target = self.material.GofR_ff_target
        self.GKofR_ff_target = self.material.GKofR_ff_target

        data = load_data(self.material.reciprocal_space_filename)
        self.q = data[:, get_index_of_function("Q", ReciprocalSpaceHeaders)]
        self.sq = data[:, get_index_of_function(
            "S(Q)", ReciprocalSpaceHeaders)]
        self.fq = data[:, get_index_of_function(
            "Q[S(Q)-1]", ReciprocalSpaceHeaders)]
        self.fq_keen = data[:, get_index_of_function(
            "FK(Q)", ReciprocalSpaceHeaders)]
        self.dcs = data[:, get_index_of_function(
            "DCS(Q)", ReciprocalSpaceHeaders)]

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.ff = FourierFilter()
        self.cutoff = 1.5

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    # Real space

    # g(r) tests

    def g_using_F(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.g_using_F(
            self.r, self.gofr, self.q, self.fq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.gofr_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def g_using_S(self):
        q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.ff.g_using_S(
            self.r, self.gofr, self.q, self.sq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.gofr_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def g_using_FK(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.g_using_FK(
            self.r, self.gofr, self.q, self.fq_keen, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.gofr_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def g_using_DCS(self):
        q_ft, dcs_ft, q, dcs, r, gr, _, _, _ = self.ff.g_using_DCS(
            self.r, self.gofr, self.q, self.dcs, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.gofr_ff_target,
                        rtol=self.rtol, atol=self.atol)

    # G(r) tests

    def G_using_F(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.G_using_F(
            self.r, self.GofR, self.q, self.fq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def G_using_S(self):
        q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.ff.G_using_S(
            self.r, self.GofR, self.q, self.sq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def G_using_FK(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.G_using_FK(
            self.r, self.GofR, self.q, self.fq_keen, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def G_using_DCS(self):
        q_ft, dcs_ft, q, dcs, r, gr, _, _, _ = self.ff.G_using_DCS(
            self.r, self.GofR, self.q, self.dcs, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    # G(r) tests

    def GK_using_F(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.GK_using_F(
            self.r, self.GKofR, self.q, self.fq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GKofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_using_S(self):
        q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.ff.GK_using_S(
            self.r, self.GKofR, self.q, self.sq, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GKofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_using_FK(self):
        q_ft, fq_ft, q, fq, r, gr, _, _, _ = self.ff.GK_using_FK(
            self.r,
            self.GKofR,
            self.q,
            self.fq_keen,
            self.cutoff,
            **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GKofR_ff_target,
                        rtol=self.rtol, atol=self.atol)

    def GK_using_DCS(self):
        q_ft, dcs_ft, q, dcs, r, gr, _, _, _ = self.ff.GK_using_DCS(
            self.r, self.GKofR, self.q, self.dcs, self.cutoff, **self.kwargs)
        first, last = self.real_space_first, self.real_space_last
        assert_allclose(gr[first:last],
                        self.GKofR_ff_target,
                        rtol=self.rtol, atol=self.atol)


class TestFourierFilterNickel(TestFourierFilterBase):
    def setUp(self):
        super(TestFourierFilterNickel, self).setUp()
        self.material = Nickel()
        self.initialize_material()

    def test_g_using_F(self):
        self.g_using_F()

    def test_g_using_S(self):
        self.g_using_S()

    def test_g_using_FK(self):
        self.g_using_FK()

    def test_g_using_DCS(self):
        self.g_using_DCS()

    def test_G_using_F(self):
        self.G_using_F()

    def test_G_using_S(self):
        self.G_using_S()

    def test_G_using_FK(self):
        self.G_using_FK()

    def test_G_using_DCS(self):
        self.G_using_DCS()

    def test_GK_using_F(self):
        self.GK_using_F()

    def test_GK_using_S(self):
        self.GK_using_S()

    def test_GK_using_FK(self):
        self.GK_using_FK()

    def test_GK_using_DCS(self):
        self.GK_using_DCS()


class TestFourierFilterArgon(TestFourierFilterBase):
    def setUp(self):
        super(TestFourierFilterArgon, self).setUp()
        self.material = Argon()
        self.initialize_material()

    def test_g_using_F(self):
        self.g_using_F()

    def test_g_using_S(self):
        self.g_using_S()

    def test_g_using_FK(self):
        self.g_using_FK()

    def test_g_using_DCS(self):
        self.g_using_DCS()

    def test_G_using_F(self):
        self.G_using_F()

    def test_G_using_S(self):
        self.G_using_S()

    def test_G_using_FK(self):
        self.G_using_FK()

    def test_G_using_DCS(self):
        self.G_using_DCS()

    def test_GK_using_F(self):
        self.GK_using_F()

    def test_GK_using_S(self):
        self.GK_using_S()

    def test_GK_using_FK(self):
        self.GK_using_FK()

    def test_GK_using_DCS(self):
        self.GK_using_DCS()


if __name__ == '__main__':
    unittest.main()  # pragma: no cover



"""
=============
StoG
=============

This module defines the StoG class
that tries to replicate the previous
stog program behavior in an organized fashion
with the ability to re-construct the workflow.
"""
from __future__ import absolute_import, division, print_function

import sys
import os
from h5py import File
import json
import numpy as np

from pystog.utils import create_domain, RealSpaceChoices, ReciprocalSpaceChoices
from pystog.converter import Converter
from pystog.transformer import Transformer
from pystog.fourier_filter import FourierFilter


class NoInputFilesException(Exception):
    """Exception when no files are given to process"""


class StoG(object):
    """
    The StoG class is used to put together
    the Converter, Transformer, and FourierFilter
    class functionalities to reproduce the original
    **stog** Fortran program behavior. This class is meant to
    put together the functionality of the classes
    into higher-level calls to construct workflows
    for merging and processing multiple recprical space
    functions into a final output real space function.

    This pythonized-version of the original Fortran **stog**
    uses numpy for data storage, organization, and
    manipulation "under the hood".

    :examples:

    >>> import json
    >>> from pystog import StoG
    >>> with open("../data/examples/argon_pystog.json", 'r') as f:
    >>>     kwargs = json.load(f)
    >>> stog = StoG(**kwargs)
    >>> stog.read_all_data()
    >>> stog.merge_data()
    >>> stog.write_out_merged_data()
    """

    def __init__(self, **kwargs):
        # General attributes
        self.__xdecimals = 2
        self.__ydecimals = 16
        self.__xmin = 100
        self.__xmax = 0
        self.__qmin = None
        self.__qmax = None
        self.__files = None
        self.__real_space_function = "g(r)"
        self.__rmin = 0.0
        self.__rmax = 50.0
        self.__rdelta = 0.01
        self.__update_dr()
        self.__density = 1.0
        self.__bcoh_sqrd = 1.0
        self.__btot_sqrd = 1.0
        self.__low_q_correction = False
        self.__lorch_flag = False
        self.__fourier_filter_cutoff = None
        self.__merged_opts = {"Y": {"Offset": 0.0, "Scale": 1.0}}
        self.__stem_name = "out"

        # Storage arrays for total scattering functions
        self.__reciprocal_individuals = np.empty([3, 0])
        self.__sq_individuals = np.empty([3, 0])
        self.__q_master = {}
        self.__sq_master = {}
        self.__r_master = {}
        self.__gr_master = {}

        # Attributes that do not (currently) change
        self.__sq_title = "S(Q) Merged"
        self.__qsq_minus_one_title = "Q[S(Q)-1] Merged"
        self._ft_title = "FT term"
        self.__sq_ft_title = "S(Q) FT"
        self.__fq_title = "F(Q) Merged"
        self.__dr_ft_title = "D(r) FT"
        self.__GKofR_title = "G(r) (Keen Version)"

        # Set real space title attributes
        self.__gr_title = "%s Merged" % self.__real_space_function
        self.__gr_ft_title = "%s FT" % self.__real_space_function
        self.__gr_lorch_title = "%s FT Lorched" % self.__real_space_function

        # Interior class attributes

        #: The **converter** attribute defines the Converter
        #: which is used to do all inner conversions necessary
        #: to go from the input reciprocal space functions,
        #: produce diagnostics for the selected real space functions,
        #: and finally output the desired real space function
        #: Must of type :class:`pystog.converter.Converter`
        self.converter = Converter()

        #: The **transformer** attribute defines the Transformer
        #: which is used to do all Fourier transforms necessary
        #: from reciprocal space to real space and vice versa.
        #: Must of type :class:`pystog.transformer. Transformer`
        self.transformer = Transformer()

        #: The **filter** attribute defines the FourierFilter
        #: which is used to do all Fourier filtering if the
        #: **fourier_filter_cutoff** attribute is supplied.
        #: Must of type :class:`pystog.fourier_filter.FourierFilter`
        self.filter = FourierFilter()

        # Use key-word arguments to set attributes
        self.__kwargs2attr(kwargs)

    def __kwargs2attr(self, kwargs):
        """
        Takes the key-word arguments supplied to the
        initialization of the class and maps them to
        attirbutes in the class. Commonly get the **kwargs**
        from the JSON input file.
        """
        if "Files" in kwargs:
            self.files = kwargs["Files"]
        if "RealSpaceFunction" in kwargs:
            self.real_space_function = str(kwargs["RealSpaceFunction"])
        if "Rmin" in kwargs:
            self.rmin = float(kwargs["Rmin"])
        if "Rmax" in kwargs:
            self.rmax = float(kwargs["Rmax"])
        if "Rdelta" in kwargs:
            self.rdelta = kwargs["Rdelta"]
        elif "Rpoints" in kwargs:
            self.rdelta = self.rmax / kwargs["Rpoints"]
        if "NumberDensity" in kwargs:
            self.density = kwargs["NumberDensity"]
        if "OmittedXrangeCorrection" in kwargs:
            self.low_q_correction = kwargs["OmittedXrangeCorrection"]
        if "LorchFlag" in kwargs:
            self.lorch_flag = kwargs["LorchFlag"]
        if "FourierFilter" in kwargs:
            if "Cutoff" in kwargs["FourierFilter"]:
                self.fourier_filter_cutoff = kwargs["FourierFilter"]["Cutoff"]
        if "<b_coh>^2" in kwargs:
            self.bcoh_sqrd = kwargs["<b_coh>^2"]
        if "<b_tot^2>" in kwargs:
            self.btot_sqrd = kwargs["<b_tot^2>"]
        if "Merging" in kwargs:
            self.merged_opts = kwargs["Merging"]
            if "Transform" in kwargs["Merging"]:
                transform_opts = kwargs["Merging"]["Transform"]
                if "Qmin" in transform_opts:
                    self.qmin = transform_opts["Qmin"]
                if "Qmax" in transform_opts:
                    self.qmax = transform_opts["Qmax"]
        if "Outputs" in kwargs:
            if "StemName" in kwargs["Outputs"]:
                self.stem_name = kwargs["Outputs"]["StemName"]
        if "NexusFile" in kwargs:
            self.nexus_file = kwargs["NexusFile"]
        if "WorkspaceName" in kwargs:
            self.workspace_name = kwargs["WorkspaceName"]

    # General attributes
    @property
    def xmin(self):
        """
        The minimum X value of all datasets to
        use for Fourier transforms (from recirocal space -> real space)

        :getter: Returns the current set value
        :setter: Set the xmin value for the Fourier transforms
        :type: float
        """
        return self.__xmin

    @xmin.setter
    def xmin(self, value):
        self.__xmin = value

    @property
    def xmax(self):
        """
        The maximum X value of all datasets to
        use for Fourier transforms (from recirocal space -> real space)

        :getter: Returns the current set value
        :setter: Set the xmax value for the Fourier transforms
        :type: float
        """
        return self.__xmax

    @xmax.setter
    def xmax(self, value):
        self.__xmax = value

    @property
    def qmin(self):
        """
        The :math:`Q_{min}` value to use for the Fourier
        transforms (from recirocal space -> real space). This
        overrides **xmin** attribute if **xmin** < **qmin**.

        :getter: Returns the current set value
        :setter: Set the :math:`Q_{min}` value for the Fourier transforms
        :type: float
        """
        return self.__qmin

    @qmin.setter
    def qmin(self, value):
        self.__qmin = value

    @property
    def qmax(self):
        """
        The :math:`Q_{max}` value to use for the Fourier
        transforms (from recirocal space -> real space). This
        overrides **xmax** attribute if **qmax** < **xmax**.

        :getter: Returns the current set value
        :setter: Set the qmax value for the Fourier transforms
        :type: float
        """
        return self.__qmax

    @qmax.setter
    def qmax(self, value):
        self.__qmax = value

    @property
    def files(self):
        """
        The files that contain the reciprocal space data
        to merge together.

        :getter: Current list of files to merge
        :setter: Set the list of files to merg
        :type: list
        """
        return self.__files

    @files.setter
    def files(self, file_list):
        self.__files = file_list

    def append_file(self, new_file):
        """
        Appends a file to the file list

        :param new_file: New file name to append
        :type new_file: str
        :return: File list with appended new_file
        :rtype: list
        """
        self.files = self.files + [new_file]
        return self.files

    def extend_file_list(self, new_files):
        """
        Extend the file list with a list of new files

        :param new_files: List of new files
        :type new_file: list
        :return: File list extended by new_files
        :rtype: list
        """
        self.files.extend(new_files)
        return self.files

    def __update_dr(self):
        """
        Uses **rdelta** and **rmax** attributes (:math:`\\Delta r` and
        :math:`R_{max}`, respectively) to construct **dr** attribute
        (:math:`r`-space vector) via its setter
        """
        self.dr = create_domain(self.rmin, self.rmax, self.rdelta)

    @property
    def rdelta(self):
        """
        The :math:`\\Delta r` for the :math:`r`-space vector

        :getter: Return :math:`\\Delta r` value
        :setter: Set the :math:`\\Delta r` value
                 and update :math:`r`-space vector
                 via the **dr** attribute
        :type: value
        """
        return self.__rdelta

    @rdelta.setter
    def rdelta(self, value):
        self.__rdelta = value
        self.__update_dr()

    @property
    def rmin(self):
        """
        The :math:`R_{min}` valuefor the :math:`r`-space vector

        :getter: Return :math:`R_{min}` value
        :setter: Set the :math:`R_{min}` value and update :math:`r`-space vector
                 via the **dr** attribute
        :type: value
        """
        return self.__rmin

    @rmin.setter
    def rmin(self, value):
        self.__rmin = value
        self.__update_dr()

    @property
    def rmax(self):
        """
        The :math:`R_{max}` valuefor the :math:`r`-space vector

        :getter: Return :math:`R_{max}` value
        :setter: Set the :math:`R_{max}` value and update :math:`r`-space vector
                 via the **dr** attribute
        :type: value
        """
        return self.__rmax

    @rmax.setter
    def rmax(self, value):
        self.__rmax = value
        self.__update_dr()

    @property
    def dr(self):
        """
        The real space function X axis data, :math:`r`-space vector

        :getter: Return the :math:`r` vector
        :setter: Set the :math:`r` vector
        :type: numpy.array
        """
        return self.__dr

    @dr.setter
    def dr(self, r):
        self.__dr = r

    @property
    def density(self):
        """
        The number density used (atoms/:math:`\\AA^{3}`) used for the
        :math:`\\rho_{0}` term in the equations

        :getter: Return the density value
        :setter: Set the density value
        :type: float
        """
        return self.__density

    @density.setter
    def density(self, value):
        self.__density = value

    @property
    def bcoh_sqrd(self):
        """
        The average coherent scattering length, squared:
        :math:`\\langle b_{coh} \\rangle^2` =
        :math:`( \\sum_{i} c_{i} b_{coh,i} ) ( \\sum_{i} c_{i} b^*_{coh,i} )`
        where the subscript :math:`i` implies for atom type :math:`i`,
        :math:`c_{i}` is the concentration of :math:`i`,
        :math:`b_{coh,i}` is the coherent scattering length of :math:`i`,
        and :math:`b^*_{coh,i}`
        is the complex coherent scattering length of :math:`i`


        The real part of the :math:`b_{coh,i}` term can be found
        from the **Coh b** column of the NIST neutron scattering
        length and cross section table found here:
        https://www.ncnr.nist.gov/resources/n-lengths/list.html

        Units are in :math:`fm` in the table for the :math:`b_{coh,i}` term.
        Thus, :math:`\\langle b_{coh} \\rangle^2` has units of :math:`fm^2`
        (and what PyStoG expects). NOTE: 100 :math:`fm^2` == 1 :math:`barn`.

        :getter: Return the value of :math:`\\langle b_{coh} \\rangle^2`
        :setter: Set the value for :math:`\\langle b_{coh} \\rangle^2`
        :type: float
        """
        # TODO: Add an example with code to demostrate
        return self.__bcoh_sqrd

    @bcoh_sqrd.setter
    def bcoh_sqrd(self, value):
        self.__bcoh_sqrd = value

    @property
    def btot_sqrd(self):
        """
        The average coherent scattering length, squared:
        :math:`\\langle b_{tot}^2 \\rangle`
        = :math:`\\sum_{i} c_{i} b_{tot,i}^2`
        = :math:`\\frac{1}{4 \\pi} \\sum_i c_{i} \\sigma_{tot,i}`
        where the subscript :math:`i` implies for atom type :math:`i`,
        :math:`c_{i}` is the concentration of :math:`i` and
        :math:`\\sigma_{tot,i}` is the total cross-section of :math:`i`


        The real part of the :math:`b_{coh,i}` term can be found
        from the **Scatt xs** column of the NIST neutron scattering
        length and cross section table found here:
        https://www.ncnr.nist.gov/resources/n-lengths/list.html

        Units are in :math:`barn` (=100 :math:`fm^2`) in the table
        for the :math:`\\sigma_{tot,i}` term. Thus, you must multiply
        :math:`\\sum_{i} c_{i} b_{tot,i}^2` by 100 to go from :math:`barn`
        to :math:`fm^2` (what PyStoG expects).

        :getter: Return the value of :math:`\\sum_{i} c_{i} b_{tot,i}^2`
        :setter: Set the value for :math:`\\sum_{i} c_{i} b_{tot,i}^2`
        :type: float
        """
        # TODO: Add an example with code to demostrate
        return self.__btot_sqrd

    @btot_sqrd.setter
    def btot_sqrd(self, value):
        self.__btot_sqrd = value

    @property
    def stem_name(self):
        """
        A stem name to prefix for all output files. Replicates
        the **stog** Fortran program behavior.

        :getter: Return the currently set stem name
        :setter: Set the stem name for output files
        :type: str
        """
        return self.__stem_name

    @stem_name.setter
    def stem_name(self, name):
        self.__stem_name = name

    @property
    def low_q_correction(self):
        """
        This sets the option to perform a low-:math:`Q` correction
        for the omitted :math:`Q` range.

        See :class:`pystog.transformer.Transformer` **_low_x_correction**
        method for more information.

        :getter: Return bool of applying the low-:math:`Q` correction
        :setter: Set whether the correction is applied or not
        :type: bool
        """
        return self.__low_q_correction

    @low_q_correction.setter
    def low_q_correction(self, value):
        if isinstance(value, bool):
            self.__low_q_correction = value
        else:
            raise TypeError("Expected a bool, True or False")

    @property
    def lorch_flag(self):
        """
        This sets the option to perform the Lorch dampening correction
        for the :math:`Q` range. Generally, will help reduce Fourier "ripples",
        or AKA "Gibbs phenomenon", due to discontinuity at :math:`Q_{max}`
        if the reciprocal space function is not at the
        :math:`Q -> \\infty` limit.
        Yet, will also broaden real space function peaks, possibly dubiously.

        See :class:`pystog.transformer.Transformer` **fourier_transform** and
        **_low_x_correction** methods for where this is applied.

        :getter: Return bool of applying the Lorch dampening correction
        :setter: Set whether the correction is applied or not
        :type: bool
        """
        return self.__lorch_flag

    @lorch_flag.setter
    def lorch_flag(self, value):
        if isinstance(value, bool):
            self.__lorch_flag = value
        else:
            raise TypeError("Expected a bool, True or False")

    @property
    def fourier_filter_cutoff(self):
        """
        This sets the cutoff in :math:`r`-space for the Fourier
        filter. The minimum is automatically 0.0. Thus, from
        0.0 to **fourier_filter_cutoff** is reverse transfomed,
        subtracted in reciprocal space, and then the difference
        is back-transformed.

        See :class:`pystog.fourier_filter.FourierFilter` for more information.

        :getter: Return currently set cutoff value
        :setter: Set cutoff value
        :type: float
        """
        return self.__fourier_filter_cutoff

    @fourier_filter_cutoff.setter
    def fourier_filter_cutoff(self, value):
        self.__fourier_filter_cutoff = value

    @property
    def merged_opts(self):
        """
        This sets the options to perform after merging the
        reciprocal space functions together, such as an
        overall offset and scale.

        :getter: Return the options currently set
        :setter: Set the options for the merged pattern
        :type: dict
        """
        return self.__merged_opts

    @merged_opts.setter
    def merged_opts(self, options):
        self.__merged_opts = options

    # Storage array attributes

    @property
    def reciprocal_individuals(self):
        """
        The storage array for the input reciprocal space functions
        loaded from **files** and with the loading processing
        from **add_dataset** class method.

        :getter: Returns the current individual, input reciprocal space
                 functions numpy array.
                 The dimenions is :math:`3 x N*M` where N is the number
                 of patterns stored and M is the length of the patterns.
        :setter: Sets the numpy array
        :type: numpy.ndarray
        """
        return self.__reciprocal_individuals

    @reciprocal_individuals.setter
    def reciprocal_individuals(self, individuals):
        self.__reciprocal_individuals = individuals

    @property
    def sq_individuals(self):
        """
        The storage array for the :math:`S(Q)` generated from each input
        reciprocal space dataset in **reciprocal_individuals** array.

        :getter: Returns the current individual :math:`S(Q)` reciprocal space
                 functions numpy array.
                 The dimenions is :math:`3 x N*M` where N is the number
                 of patterns stored and M is the length of the patterns.
        :setter: Sets the numpy array
        :type: numpy.ndarray
        """
        return self.__sq_individuals

    @sq_individuals.setter
    def sq_individuals(self, individuals):
        self.__sq_individuals = individuals

    @property
    def sq_master(self):
        """
        The "master" dictionary for the :math:`S(Q)` reciprocal
        space functions that are generated for each processing step.

        :getter: Returns the current "master" :math:`S(Q)` reciprocal space
                 functions dictionary generated up to the current step in
                 the workflow.
        :setter: Sets the "master" :math:`S(Q)` function dictionary
        :type: dict[str:numpy.ndarray]
        """
        return self.__sq_master

    @sq_master.setter
    def sq_master(self, sq):
        self.__sq_master = sq

    @property
    def gr_master(self):
        """
        The "master" dictionary for the real space functions
        that are generated for each processing step.

        :getter: Returns the current "master" real space
                 functions dictionary generated up to the current step in
                 the workflow.
        :setter: Sets the "master" real space function dictionary
        :type: dict[str:numpy.ndarray]
        """
        return self.__gr_master

    @gr_master.setter
    def gr_master(self, gr):
        self.__gr_master = gr

    @property
    def q_master(self):
        """
        The "master" dictionary for the domain :math:Q` of the reciprocal
        space functions that are generated for each processing step.

        :getter: Returns the current "master" :math:`Q` reciprocal space
                 functions dictionary generated up to the current step in
                 the workflow.
        :setter: Sets the "master" :math:`Q` dictionary
        :type: dict[str:numpy.ndarray]
        """
        return self.__q_master

    @q_master.setter
    def q_master(self, q):
        self.__q_master = q

    @property
    def r_master(self):
        """
        The "master" dictionary for the domain :math:r` of the real
        space functions that are generated for each processing step.

        :getter: Returns the current "master" :math:`r` real space
                 functions dictionary generated up to the current step in
                 the workflow.
        :setter: Sets the "master" :math:`r` dictionary
        :type: dict[str:numpy.ndarray]
        """
        return self.__r_master

    @r_master.setter
    def r_master(self, r):
        self.__r_master = r

    @property
    def real_space_function(self):
        """
        The real space function to use throughoutt the processing

        :getter: Returns the currently select real space function
        :setter: Set the selected real space function and
                 updates other title attributes that rely on this
                 in their name.
        :type: str
        """
        return self.__real_space_function

    @real_space_function.setter
    def real_space_function(self, real_space_function):
        if real_space_function not in RealSpaceChoices:
            raise ValueError(
                "real_space_function must be of %s"
                % ",".join(RealSpaceChoices.keys())
            )
        self.__real_space_function = real_space_function
        self.__gr_ft_title = "%s FT" % real_space_function
        self.__gr_lorch_title = "%s FT Lorched" % real_space_function
        self.__gr_title = "%s Merged" % real_space_function

    @property
    def sq_title(self):
        """
        The title of the :math:`S(Q)` function directly after merging
        the reciprocal space functions without any further corrections.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__sq_title

    @sq_title.setter
    def sq_title(self, title):
        self.__sq_title = title

    @property
    def qsq_minus_one_title(self):
        """
        The title of the :math:`Q[S(Q)-1]` function
        directly after merging the reciprocal space
        functions without any further corrections.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__qsq_minus_one_title

    @qsq_minus_one_title.setter
    def qsq_minus_one_title(self, title):
        self.__qsq_minus_one_title = title

    @property
    def sq_ft_title(self):
        """
        The title of the :math:`S(Q)` function after
        merging and a fourier filter correction.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__sq_ft_title

    @sq_ft_title.setter
    def sq_ft_title(self, title):
        self.__sq_ft_title = title

    @property
    def fq_title(self):
        """
        The title of the :math:`F(Q)` function after
        merging and a fourier filter correction.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__fq_title

    @fq_title.setter
    def fq_title(self, title):
        self.__fq_title = title

    @property
    def gr_title(self):
        """
        The title of the real space function directly after merging
        the reciprocal space functions without any further corrections.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__gr_title

    @gr_title.setter
    def gr_title(self, title):
        self.__gr_title = title

    @property
    def gr_ft_title(self):
        """
        The title for the real space function after both
        merging and a fourier filter correction

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__gr_ft_title

    @gr_ft_title.setter
    def gr_ft_title(self, title):
        self.__gr_ft_title = title

    @property
    def gr_lorch_title(self):
        """
        The title for the real space function with the lorch correction

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__gr_lorch_title

    @gr_lorch_title.setter
    def gr_lorch_title(self, title):
        self.__gr_lorch_title = title

    @property
    def GKofR_title(self):
        """
        The title of the :math:`G_{Keen Version}(r)` with
        all corrections applied.

        :getter: Returns the current title for this function
        :setter: Sets the title for this function
        :type: str
        """
        return self.__GKofR_title

    @GKofR_title.setter
    def GKofR_title(self, title):
        self.__GKofR_title = title

    # -------------------------------------#
    # Reading and Merging Spectrum

    def extract(self, hdf_file, path, index=None):
        data = File(hdf_file, "r")
        # base = path.split("/")[1]
        # print(data[base+"/title"].value)
        if index is not None:
            return data[path][index]
        return data[path][()]

    def extract_path_from_title(
        self, hdf_file, title, title_path="title", wksp_path="workspace"
    ):
        data = File(hdf_file, "r")
        choices = list()
        for name, group in data.items():
            index = os.path.join(name, title_path)
            if data[index][(0)].decode("UTF-8") == title:
                xpath = os.path.join("/", name, wksp_path, "axis1")
                ypath = os.path.join("/", name, wksp_path, "values")
                return xpath, ypath
            else:
                choices.append(data[index][()])

        print("ERROR: Did not find a workspace with title: %s" % title)
        print("       These are the workspaces titles found in this file")
        for c in sorted(choices):
            print("      %s" % c)
        sys.exit("Stopping...")

    def extract_xy(self, hdf_file, xpath, ypath, **kwargs):
        x = self.extract(hdf_file, xpath)
        y = self.extract(hdf_file, ypath, **kwargs)
        y = np.insert(y, 0, y[0])  # hack for histogram xaxis
        return x, y

    def save_xy(self, filename, xdata, ydata):
        # assert(len(xdata) == len(ydata))
        with open(filename, "w") as f:
            f.write("%d\n\n" % len(xdata))
            for x, y in zip(xdata, ydata):
                f.write("%f %f\n" % (x, y))

    def read_nexus_file_by_bank(self, nexus_file, bank, title, **kwargs):
        """
        Reads an individual file bank by bank and uses the **extract_xy**
        to handle extraction and **extract_path_from_title** to obtain
        coord paths
        """

        xpath, ypath = self.extract_path_from_title(nexus_file, title)
        output_file = "{}_bank{}.dat".format(title, bank)

        output_file = self.stem_name + output_file

        x, y = self.extract_xy(nexus_file, xpath, ypath, index=bank)
        self.save_xy(output_file, x, y)

    def read_all_nexus_file_banks(self):
        # Check that we have files to operate on
        if not self.files:
            raise NoInputFilesException("No input files given in arguments")
        if not self.nexus_file:
            raise NoInputFilesException("No nexus file given in arguments")
        if not self.workspace_name:
            raise NoInputFilesException("No workspace name given in arguments")

        # Read in all the data files
        for i, file_info in enumerate(self.files):
            self.read_nexus_file_by_bank(
                self.nexus_file,
                int(file_info["BankNumber"]),
                self.workspace_name,
            )

    def read_all_data(self, **kwargs):
        """
        Reads all the data from the **files** attribute
        Uses the **read_dataset** method on each file.

        Will append all datasets to the numpy storage array,
        **reciprocal_individuals**, and also convert to :math:`S(Q)` and add to
        the **sq_individuals** numpy storage array in **add_dataset** method
        via **read_dataset** method.
        """
        # Check that we have files to operate on
        if not self.files:
            raise NoInputFilesException("No input files given in arguments")

        # Read in all the data files
        for i, file_info in enumerate(self.files):
            self.read_dataset(file_info, **kwargs)

    def read_dataset(
        self, info, xcol=0, ycol=1, dycol=2, sep=r"\s+", skiprows=2, **kwargs
    ):
        """
        Reads an individual file and uses the **add_dataset**
        method to apply all dataset manipulations, such as
        scales, offsets, cropping, etc.

        :param info: Dict with information for dataset
                     (filename, manipulations, etc.)
        :type info: dict
        :param xcol: Column in data file for X-axis
        :type xcol: int
        :param ycol: Column in data file for Y-axis
        :type ycol: int
        :param dycol: Column in data file for Y uncertainty
        :type dycol: int
        :param sep: Separator for the file used by numpy.loadtxt
        :type sep: raw string
        :param skiprows: Number of rows to skip. Passed to numpy.loadtxt
        :type skiprows: int
        """
        _data = np.loadtxt(
            info["Filename"], skiprows=skiprows, comments="#", unpack=True
        )
        if _data.shape[0] <= xcol or _data.shape[0] <= ycol:
            raise RuntimeError("Data format incompatible with input parameters")
        if _data.shape[0] <= dycol:
            array_seq = (_data[xcol], _data[ycol], np.zeros_like(_data[ycol]))
            data = np.stack(array_seq)
        else:
            data = np.stack((_data[xcol], _data[ycol], _data[dycol]))
        info["data"] = data
        self.add_dataset(info, **kwargs)

    def add_dataset(
        self, info, yscale=1.0, yoffset=0.0, xoffset=0.0, ydecimals=16, **kwargs
    ):
        """
        Takes the info with the dataset and manipulations,
        such as scales, offsets, cropping, etc., and creates
        an invidual numpy array for the pattern.

        :param info: Dict with information for dataset
                     (filename, manipulations, etc.)
        :type info: dict
        :param yscale: Scale factor for the Y data
                       (i.e. :math:`S(Q)`, :math:`F(Q)`, etc.)
        :type yscale: float
        :param yoffset: Offset factor for the Y data
                        (i.e. :math:`S(Q)`, :math:`F(Q)`, etc.)
        :type yoffset: float
        :param xoffset: Offset factor for the X data (i.e. :math:`Q`)
        :type yoffset: float
        """
        # Extract data
        x = np.around(np.array(info["data"][0]), decimals=self.__xdecimals)
        y = np.around(np.array(info["data"][1]), decimals=self.__ydecimals)
        if len(info["data"]) == 3:
            dy = np.around(np.array(info["data"][2]), decimals=self.__ydecimals)
        else:
            dy = np.zeros_like(y)

        # Cropping
        xmin = min(x)
        xmax = max(x)
        if "Qmin" in info:
            xmin = info["Qmin"]
        if "Qmax" in info:
            xmax = info["Qmax"]
        x, y, dy = self.transformer.apply_cropping(x, y, xmin, xmax, dy=dy)

        # Offset and scale
        adjusting = False
        if "Y" in info:
            adjusting = True
            if "Scale" in info["Y"]:
                yscale = info["Y"]["Scale"]
            if "Offset" in info["Y"]:
                yoffset = info["Y"]["Offset"]

        if "X" in info:
            adjusting = True
            if "Offset" in info["X"]:
                xoffset = info["X"]["Offset"]

        if adjusting:
            x, y, dy = self.apply_scales_and_offset(
                x, y, dy=dy, yscale=yscale, yoffset=yoffset, xoffset=xoffset
            )

        # Save overal x-axis min and max
        self.xmin = min(self.xmin, xmin)
        self.xmax = max(self.xmax, xmax)

        # Use Qmin and Qmax to crop datasets
        if self.qmin is not None:
            if self.xmin < self.qmin:
                x, y, dy = self.transformer.apply_cropping(
                    x, y, self.qmin, self.xmax, dy=dy
                )
        if self.qmax is not None:
            if self.xmax > self.qmax:
                x, y, dy = self.transformer.apply_cropping(
                    x, y, self.xmin, self.qmax, dy=dy
                )

        # Default to S(Q) if function type not defined
        if "ReciprocalFunction" not in info:
            info["ReciprocalFunction"] = "S(Q)"

        if info["ReciprocalFunction"] not in ReciprocalSpaceChoices:
            error = "ReciprocalFunction was equal to {given}.\n"
            error += "ReciprocalFunction must be one of the folloing {options}"
            error = error.format(
                given=info["ReciprocalFunction"],
                options=json.dumps(ReciprocalSpaceChoices),
            )
            raise ValueError(error)

        # Save reciprocal space function to the "invididuals" array
        array_seq = (self.reciprocal_individuals, np.stack((x, y, dy)))
        self.reciprocal_individuals = np.concatenate(array_seq, axis=1)

        # Convert to S(Q) and save to the individual S(Q) array
        if info["ReciprocalFunction"] == "Q[S(Q)-1]":
            y, dy = self.converter.F_to_S(x, y, dfq=dy)
        elif info["ReciprocalFunction"] == "FK(Q)":
            y, dy = self.converter.FK_to_S(
                x, y, dfq_keen=dy, **{"<b_coh>^2": self.bcoh_sqrd}
            )
        elif info["ReciprocalFunction"] == "DCS(Q)":
            y, dy = self.converter.DCS_to_S(
                x,
                y,
                ddcs=dy,
                **{"<b_coh>^2": self.bcoh_sqrd, "<b_tot^2>": self.btot_sqrd}
            )
        array_seq = (self.sq_individuals, np.stack((x, y, dy)))
        self.sq_individuals = np.concatenate(array_seq, axis=1)

    @staticmethod
    def apply_scales_and_offset(
        x, y, dy=None, yscale=1.0, yoffset=0.0, xoffset=0.0
    ):
        """
        Applies scales to the Y-axis and offsets to both X and Y axes.

        :param x: X-axis data
        :type x: numpy.array or list
        :param y: Y-axis data
        :type y: numpy.array or list
        :param yscale: Y-axis scale factor
        :type yscale: float
        :param yoffset: Y-axis offset factor
        :type yoffset: float
        :param xoffset: X-axis offset factor
        :type xoffset: float
        :return: X and Y vectors after scales and offsets applied
        :rtype: numpy.array pair
        """
        y = y * yscale
        y = y + yoffset
        x = x + xoffset
        if dy is None:
            dy = np.zeros_like(y)
        dy = dy * yscale
        return x, y, dy

    def merge_data(self):
        """
        Merges the reciprocal space data stored in the
        **reciprocal_individuals** numpy array into a single, merged
        recirocal space function. Stores the S(Q) result in
        **sq_master** dictionary using **sq_title** (default: "S(Q) Merged").

        Also, converts this
        merged :math:`S(Q)` into :math:`Q[S(Q)-1]` via the
        **Converter** class and applies any modification
        specified in **merged_opts** dict attribute, specified
        by the **'Q[S(Q)-1]'** key of the dict. If there is modification,
        this modified :math:`Q[S(Q)-1]` will be converted to
        :math:`S(Q)` and replace the :math:`S(Q)` directly after merge.

        Example dict of **merged_opts** for scaling of
        :math:`S(Q)` by 2 and then offsetting :math:`Q[S(Q)-1]` by 5:

        .. highlight:: python
        .. code-block:: python

                {
                    "Merging": {
                        "Y": {
                            "Offset": 0.0,
                            "Scale": 2.0
                        },
                        "Q[S(Q)-1]": {
                            "Y": "Offset": 5.0,
                            "Scale": 1.0
                        }
                    }
                }
        ...
        """
        data_merged = []
        _n_total = 0
        _n_sum = 0
        _n_err = 0
        _previous_x = None

        # At this point we have a concatenated array of data
        # We need to sort according to Q first
        # TODO: this is ugly but works for now.
        _x = self.sq_individuals[0]
        _y = self.sq_individuals[1]
        _dy = self.sq_individuals[2]
        zipped = zip(_x, _y, _dy)
        ordered = sorted(zipped, key=lambda a: a[0])
        _x, _y, _dy = zip(*ordered)
        self.sq_individuals = np.stack(
            (np.asarray(_x), np.asarray(_y), np.asarray(_dy))
        )

        # Go through the data and compute the average of points
        # with the same Q.
        for item in self.sq_individuals.T:
            if item[0] == _previous_x:
                _n_total += 1.0
                _n_sum += item[1]
                _n_err += item[2] ** 2
            else:
                if _n_total > 0:
                    data_merged.append(
                        [
                            _previous_x,
                            _n_sum / _n_total,
                            np.sqrt(_n_err) / _n_total,
                        ]
                    )
                _n_total = 1.0
                _n_sum = item[1]
                _n_err = item[2] ** 2
            _previous_x = item[0]
        if _n_total > 0:
            data_merged.append(
                [_previous_x, _n_sum / _n_total, np.sqrt(_n_err) / _n_total]
            )

        data_merged = np.asarray(data_merged).T

        q = data_merged[0]
        sq = data_merged[1]
        dsq = data_merged[2]

        q, sq, dsq = self.apply_scales_and_offset(
            q,
            sq,
            yscale=self.merged_opts["Y"]["Scale"],
            yoffset=self.merged_opts["Y"]["Offset"],
            xoffset=0.0,
            dy=dsq,
        )
        self.q_master[self.sq_title] = q
        self.sq_master[self.sq_title] = sq

        # Also, create merged Q[S(Q)-1] with modifications, if specified
        fofq, dfofq = self.converter.S_to_F(q, sq, dsq=dsq)
        if "Q[S(Q)-1]" in self.merged_opts:
            fofq_opts = self.merged_opts["Q[S(Q)-1]"]
            if "Y" in fofq_opts:
                if "Scale" in fofq_opts["Y"]:
                    fofq *= fofq_opts["Y"]["Scale"]
                if "Offset" in fofq_opts["Y"]:
                    fofq += fofq_opts["Y"]["Offset"]
        self.q_master[self.qsq_minus_one_title] = q
        self.sq_master[self.qsq_minus_one_title] = fofq

        # Convert this Q[S(Q)-1] back to S(Q) and overwrite the 1st one
        sq, dsq = self.converter.F_to_S(q, fofq, dfq=dfofq)
        sq[np.isnan(sq)] = 0
        self.sq_master[self.sq_title] = sq

    # -------------------------------------#
    # Transform Utilities

    def transform_merged(self):
        """
        Performs the Fourier transform on the merged **sq_master**
        pattern to generate the desired real space function
        with this correction. The results for real space are:
        the domain is saved to the **r_master** dictionary and
        the range is saved to the **gr_master** dictionary,
        with both using the **gr_title** for the key of the dictionaries.
        """
        # Create r-space vector if needed
        if self.dr is None or len(self.dr) == 0:
            self.__update_dr()

        # Get Q and S(Q)
        q = self.q_master[self.sq_title]
        sq = self.sq_master[self.sq_title]

        # Perform the Fourier transform to selected real space function
        transform_kwargs = {
            "lorch": False,
            "rho": self.density,
            "<b_coh>^2": self.bcoh_sqrd,
        }
        if self.real_space_function == "g(r)":
            r, gofr, dgofr = self.transformer.S_to_g(
                q, sq, self.dr, **transform_kwargs
            )
        elif self.real_space_function == "G(r)":
            r, gofr, dgofr = self.transformer.S_to_G(
                q, sq, self.dr, **transform_kwargs
            )
        elif self.real_space_function == "GK(r)":
            r, gofr, dgofr = self.transformer.S_to_GK(
                q, sq, self.dr, **transform_kwargs
            )

        self.gr_master[self.gr_title] = gofr
        self.r_master[self.gr_title] = r

    def fourier_filter(self):
        """
        Performs the Fourier filter on the **sq_master**
        pattern to generate the desired real space function
        with this correction. The results from both reciprocal space and
        real space are:

        1. Saved back to the respective "master" dictionaries
        2. Saved to files via the **stem_name**
        3. Returned from function

        :return: Returns a tuple with :math:`r`,
                 the selected real space function,
                 :math:`Q`,
                 and :math:`S(Q)` functions
        :rtype: tuple of numpy.array
        """
        kwargs = {
            "lorch": False,
            "rho": self.density,
            "<b_coh>^2": self.bcoh_sqrd,
            "OmittedXrangeCorrection": self.low_q_correction,
        }
        cutoff = self.fourier_filter_cutoff

        # Get reciprocal and real space data
        if self.gr_title not in self.gr_master:
            msg = (
                "WARNING: Fourier filtered before initial transform. "
                "Peforming now..."
            )
            print(msg)
            self.transform_merged()

        r = self.r_master[self.gr_title]
        gr = self.gr_master[self.gr_title]
        q = self.q_master[self.sq_title]
        sq = self.sq_master[self.sq_title]

        # Fourier filter g(r)
        # NOTE: Real space function setter will catch ValueError so
        # so no need for `else` to catch error
        if self.real_space_function == "g(r)":
            q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.filter.g_using_S(
                r, gr, q, sq, cutoff, **kwargs
            )
        elif self.real_space_function == "G(r)":
            q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.filter.G_using_S(
                r, gr, q, sq, cutoff, **kwargs
            )
        elif self.real_space_function == "GK(r)":
            q_ft, sq_ft, q, sq, r, gr, _, _, _ = self.filter.GK_using_S(
                r, gr, q, sq, cutoff, **kwargs
            )

        # Round to avoid mismatch index in domain and NaN
        q = np.around(q, decimals=self.__xdecimals)
        sq = np.around(sq, decimals=self.__ydecimals)
        q_ft = np.around(q_ft, decimals=self.__xdecimals)
        sq_ft = np.around(sq_ft, decimals=self.__ydecimals)

        # Add output to master dataframes and write files
        self.q_master[self._ft_title] = q_ft
        self.sq_master[self._ft_title] = sq_ft
        self.write_out_ft()

        self.q_master[self.sq_ft_title] = q
        self.sq_master[self.sq_ft_title] = sq
        self.write_out_ft_sq()

        self.r_master[self.gr_ft_title] = r
        self.gr_master[self.gr_ft_title] = gr
        self.write_out_ft_gr()

        return q, sq, r, gr

    def apply_lorch(self, q, sq, r):
        """
        Performs the Fourier transform using the Lorch
        dampening correction on the merged :math:`S(Q)` from
        the **sq_master** dictionary to generate the
        desired real space function with
        this correction. The results from both reciprocal space and
        real space are:

        1. Saved back to the respective "master" dictionaries
        2. Saved to files via the **stem_name**
        3. Returned from function

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :return: Returns a tuple with :math:`r` and selected real space function
        :rtype: tuple of numpy.array
        """
        if self.real_space_function == "g(r)":
            r, gr_lorch, _ = self.transformer.S_to_g(
                q, sq, r, **{"lorch": True, "rho": self.density}
            )
        elif self.real_space_function == "G(r)":
            r, gr_lorch, _ = self.transformer.S_to_G(
                q, sq, r, **{"lorch": True}
            )
        elif self.real_space_function == "GK(r)":
            r, gr_lorch, _ = self.transformer.S_to_GK(
                q,
                sq,
                r,
                **{
                    "lorch": True,
                    "rho": self.density,
                    "<b_coh>^2": self.bcoh_sqrd,
                }
            )

        self.gr_master[self.gr_lorch_title] = gr_lorch
        self.r_master[self.gr_lorch_title] = r
        self.write_out_lorched_gr()

        return r, gr_lorch

    def _get_lowR_mean_square(self):
        """
        Retuns the low-R mean square value for the real space function stored
        in the "master" real space function, **gr_master**.
        Used as a cost function for optimiziation of the :math:`Q_{max}` value
        by an iterative adjustment. Calls **_lowR_mean_square* method.
        **Currently not used in PyStoG workflow since was done manually.**

        :return: The calculated low-R mean-square value
        :rtype: float
        """
        # TODO: Automate the :math:`Q_{max}` adjustment in an iterative loop
        # using a minimizer.
        gr = self.gr_master[self.gr_title]
        return self._lowR_mean_square(self.dr, gr)

    def _lowR_mean_square(self, r, gr, limit=1.01):
        """
        Calculates the low-R mean square value from a given real space function.
        Used as a cost function for optimiziation of the :math:`Q_{max}` value
        by an iterative adjustment.
        **Currently not used in PyStoG workflow since was done manually.**

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: real space function vector
        :type gr: numpy.array or list
        :param limit: The upper limit on :math:`r` to use for
                      the mean-square calculation
        :type limit: float
        :return: The calculated low-R mean-square value
        :rtype: float
        """
        # TODO: Automate the :math:`Q_{max}` adjustment in an iterative loop
        # using a minimizer.
        gr = gr[r <= limit]
        gr_sq = np.multiply(gr, gr)
        average = sum(gr_sq)
        return np.sqrt(average)

    def _add_keen_fq(self, q, sq):
        """
        Adds the Keen version of :math:`F(Q)` to the
        "master" recprical space storage array, **sq_master**, and
        writes it out to file using the **stem_name**.

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        """
        kwargs = {"rho": self.density, "<b_coh>^2": self.bcoh_sqrd}
        fq, dfq = self.converter.S_to_FK(q, sq, **kwargs)
        self.sq_master[self.fq_title] = fq
        self.q_master[self.fq_title] = q
        self.write_out_rmc_fq()

    def _add_keen_gr(self, r, gr):
        """
        Adds the Keen version of :math:`G(r)` to the
        "master" real space storage array, **gr_master**, and
        writes it out to file using the **stem_name**.

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: real space function vector
        :type gr: numpy.array or list
        """
        kwargs = {"rho": self.density, "<b_coh>^2": self.bcoh_sqrd}
        if self.real_space_function == "g(r)":
            GKofR, dGKofR = self.converter.g_to_GK(r, gr, **kwargs)
        elif self.real_space_function == "G(r)":
            GKofR, dGKofR = self.converter.G_to_GK(r, gr, **kwargs)
        elif self.real_space_function == "GK(r)":
            GKofR = gr

        self.gr_master[self.GKofR_title] = GKofR
        self.r_master[self.GKofR_title] = r
        self.write_out_rmc_gr()

    # -------------------------------------#
    # Output Utilities
    def _write_out_to_file(self, x, y, filename, places=12):
        """
        Helper function for writing out X Y data
        to the filename in the RMCProfile format.

        :param x: X data to write out
        :type x: list
        :param y: Y data to write out
        :type y: list
        :param filename: Filename to write to
        :type filename: str
        """
        with open(filename, "w") as f:
            f.write("%d \n" % len(x))
            f.write("# Comment line\n")
        with open(filename, "a") as f:
            for i, j in zip(x, y):
                fmt = "{:.{places}f} {:.{places}f}\n"
                f.write(fmt.format(i, j, places=places))

    def write_out_merged_sq(self, filename=None):
        """
        Helper function for writing out the merged :math:`S(Q)`

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s.sq" % self.stem_name
        x = self.q_master[self.sq_title]
        y = self.sq_master[self.sq_title]
        self._write_out_to_file(x, y, filename)

    def write_out_merged_gr(self, filename=None):
        """
        Helper function for writing out the merged real space function

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s.gr" % self.stem_name
        x = self.r_master[self.gr_title]
        y = self.gr_master[self.gr_title]
        self._write_out_to_file(x, y, filename)

    def write_out_ft(self, filename=None):
        """
        Helper function for writing out the Fourier filter correction.

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "ft.dat"
        x = self.q_master[self._ft_title]
        y = self.sq_master[self._ft_title]
        self._write_out_to_file(x, y, filename)

    def write_out_ft_sq(self, filename=None):
        """
        Helper function for writing out the Fourier filtered :math:`S(Q)`

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s_ft.sq" % self.stem_name
        x = self.q_master[self.sq_ft_title]
        y = self.sq_master[self.sq_ft_title]
        self._write_out_to_file(x, y, filename)

    def write_out_ft_gr(self, filename=None):
        """
        Helper function for writing out the Fourier filtered real space function

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s_ft.gr" % self.stem_name
        x = self.r_master[self.gr_ft_title]
        y = self.gr_master[self.gr_ft_title]
        self._write_out_to_file(x, y, filename)

    def write_out_lorched_gr(self, filename=None):
        """
        Helper function for writing out the Lorch dampened real space function

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s_ft_lorched.gr" % self.stem_name
        x = self.r_master[self.gr_lorch_title]
        y = self.gr_master[self.gr_lorch_title]
        self._write_out_to_file(x, y, filename)

    def write_out_rmc_fq(self, filename=None):
        """
        Helper function for writing out the output :math:`F(Q)`

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s_rmc.fq" % self.stem_name
        x = self.q_master[self.fq_title]
        y = self.sq_master[self.fq_title]
        self._write_out_to_file(x, y, filename)

    def write_out_rmc_gr(self, filename=None):
        """
        Helper function for writing out the output :math:`G_{Keen Version}(Q)`

        :param filename: Filename to write to
        :type filename: str
        """
        if filename is None:
            filename = "%s_rmc.gr" % self.stem_name
        x = self.r_master[self.GKofR_title]
        y = self.gr_master[self.GKofR_title]
        self._write_out_to_file(x, y, filename)



import argparse
import pytest
from pystog.io import get_cli_parser, parse_cli_args


def test_get_cli_parser():
    parser = get_cli_parser()
    args = parser.parse_args([])
    assert not args.json
    assert not args.density
    assert not args.filenames
    assert args.stem_name == 'merged'
    assert args.real_space_function == 'g(r)'
    assert args.Rmax == 50.0
    assert args.Rpoints == 5000
    assert not args.Rdelta
    assert not args.fourier_filter_cutoff
    assert args.lorch_flag is False
    assert args.bcoh_sqrd == 1.0
    assert args.btot_sqrd == 1.0
    assert args.merging == [0.0, 1.0]
    assert args.low_q_correction is False


def test_parse_cli_args():
    args = argparse.Namespace()
    file0 = ["a.txt", 0.0, 30.0, 0.0, 1.0, 0.0, "S(Q)"]
    file1 = ["b.txt", 5.0, 20.0, 2.0, 3.0, 2.0, "F(Q)"]
    args.filenames = [
        file0,
        file1,
    ]
    args.density = 0.5
    args.Rmax = 100.0
    args.Rpoints = 10000
    args.fourier_filter_cutoff = 0.0
    args.lorch_flag = False
    args.stem_name = "deleteme"
    args.merging = [0.0, 1.0]
    args.bcoh_sqrd = 22.0
    args.btot_sqrd = 22.1
    args.real_space_function = "g(r)"
    args.low_q_correction = False
    args.Rdelta = 0.01
    kwargs = parse_cli_args(args)

    for i, f in enumerate([file0, file1]):
        assert kwargs["Files"][i]["Filename"] == f[0]
        assert kwargs["Files"][i]["Qmin"] == f[1]
        assert kwargs["Files"][i]["Qmax"] == f[2]
        assert kwargs["Files"][i]["Y"]["Offset"] == f[3]
        assert kwargs["Files"][i]["Y"]["Scale"] == f[4]
        assert kwargs["Files"][i]["X"]["Offset"] == f[5]
        assert kwargs["Files"][i]["ReciprocalFunction"] == f[6]
    assert kwargs["NumberDensity"] == args.density
    assert kwargs["Rmax"] == args.Rmax
    assert kwargs["Rpoints"] == args.Rpoints
    assert kwargs["FourierFilter"]["Cutoff"] == args.fourier_filter_cutoff
    assert kwargs["LorchFlag"] == args.lorch_flag
    assert kwargs["Outputs"]["StemName"] == args.stem_name
    assert kwargs["Merging"]["Y"]["Offset"] == args.merging[0]
    assert kwargs["Merging"]["Y"]["Scale"] == args.merging[1]
    assert kwargs["<b_coh>^2"] == args.bcoh_sqrd
    assert kwargs["<b_tot^2>"] == args.btot_sqrd
    assert kwargs["RealSpaceFunction"] == args.real_space_function
    assert kwargs["OmittedXrangeCorrection"] == args.low_q_correction
    assert kwargs["Rdelta"] == args.Rdelta


def test_parse_cli_args_exception():
    with pytest.raises(TypeError):
        parse_cli_args([])



"""
=============
Transformer
=============

This module defines the Transformer class
that performs the Fourier transforms
"""

import numpy as np

from pystog.converter import Converter

# -------------------------------------------------------#
# Transforms between Reciprocal and Real Space Functions


class Transformer:
    """
    The Transformer class is used to Fourier transform
    between the difference spaces. Either:
    a reciprocal space function -> real space function
    or a real space function -> reciprocal space function

    :examples:

    >>> import numpy
    >>> from pystog import Transformer
    >>> transformer = Transformer()
    >>> q, sq = numpy.loadtxt("my_sofq_file.txt",unpack=True)
    >>> r = numpy.linspace(0., 25., 2500)
    >>> r, gr, dgr = transformer.S_to_G(q, sq, r)
    >>> q = numpy.linspace(0., 25., 2500)
    >>> q, sq, dsq = transformer.G_to_S(r, gr, q)
    """

    def __init__(self):
        self.converter = Converter()

    def _low_x_correction(self, xin, yin, xout, yout, **kwargs):
        """
        Omitted low-x range correction performed in the
        space you have transformed to. Does so by assumming a
        linear extrapolation to zero.
        Original author: Jack Carpenter

        :param xin: domain vector for space to be transformed from
        :type xin: numpy.array or list
        :param yin: range vector for space to be transformed from
        :type yin: numpy.array or list
        :param xout: domain vector for space to be transformed to
        :type xout: numpy.array or list
        :param yin: range vector for space to be transformed to
        :type yin: numpy.array or list
        :return: range vector for space transformed to with correction applied
        :rtype: numpy.array
        """
        """
        TODO: Refactor correction to just
        1) peform linear extrapolation in space before transform
        2) transform with extrapolated function
        Compare that implementation to this one.
        Replace this if equal (within tolerance)
        """

        lorch_flag = kwargs.get('lorch', False)

        xmin = min(xin)
        xmax = max(xin)
        yin_xmin = yin[0]
        np.pi / xmax

        PiOverXmax = np.pi / xmax

        correction = np.zeros_like(yout)
        for i, x in enumerate(xout):
            v = xmin * x
            if lorch_flag:
                vm = xmin * (x - PiOverXmax)
                vp = xmin * (x + PiOverXmax)
                term1 = (vm * np.sin(vm) + np.cos(vm) - 1.) / \
                    (x - PiOverXmax)**2.
                term2 = (vp * np.sin(vp) + np.cos(vp) - 1.) / \
                    (x + PiOverXmax)**2.
                F1 = (term1 - term2) / (2. * PiOverXmax)
                F2 = (np.sin(vm) / (x - PiOverXmax) - np.sin(vp) / # noqa
                      (x + PiOverXmax)) / (2. * PiOverXmax)
            else:
                F1 = (2. * v * np.sin(v) - (v * v - 2.) * np.cos(v) - 2.)
                F1 = np.divide(
                    F1, x * x * x, out=np.zeros_like(F1), where=x != 0)

                F2 = (np.sin(v) - v * np.cos(v))
                F2 = np.divide(F2, x * x, out=np.zeros_like(F2), where=x != 0)

            num = F1 * yin_xmin
            factor = np.divide(
                num, xmin, out=np.zeros_like(num), where=xmin != 0)
            correction[i] = (2 / np.pi) * (factor - F2)

        yout += correction

        return yout

    def apply_cropping(self, x, y, xmin, xmax, dy=None):
        """
        Utility to crop x and y based on xmin and xmax along x.
        Provides the capability to specify the (Qmin,Qmax)
        or (Rmin,Rmax) in the Fourier transform

        :param x: domain vector
        :type x: numpy.array or list
        :param y: range vector
        :type y: numpy.array or list
        :param xmin: minimum x-value for crop
        :type xmin: float
        :param xmax: maximum x-value for crop
        :type xmax: float
        :param dy: uncertainty vector
        :type dy: numpy.array or list
        :return: vector pair (x,y) with cropping applied
        :rtype: (numpy.array, numpy.array, numpy.array)
        """
        if dy is not None:
            err = np.asarray(dy)
        else:
            err = np.zeros_like(y)
        indices = np.logical_and(x >= xmin, x <= xmax)
        return x[indices], y[indices], err[indices]

    def fourier_transform(self,
                          xin,
                          yin,
                          xout,
                          xmin=None,
                          xmax=None,
                          dyin=None,
                          **kwargs):
        """
        The Fourier transform function. The kwargs
        argument allows for different modifications:
        Lorch dampening, omitted low-x range correction,

        :param xin: domain vector for space to be transformed from
        :type xin: numpy.array or list
        :param yin: range vector for space to be transformed from
        :type yin: numpy.array or list
        :param xout: domain vector for space to be transformed to
        :type xout: numpy.array or list
        :param xmin: minimum x-value for crop
        :type xmin: float
        :param xmax: maximum x-value for crop
        :type xmax: float
        :param dyin: uncertainty vector for yin
        :type dyin: numpy.array or list
        :return: vector pair of transformed domain, range vectors,
                 and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """

        if xmax is None:
            xmax = max(xin)
        if xmin is None:
            xmin = min(xin)

        xin, yin, err = self.apply_cropping(xin, yin, xmin, xmax, dyin)

        factor = np.ones_like(yin)
        if kwargs.get('lorch', False):
            PiOverXmax = np.pi / xmax
            num = np.sin(PiOverXmax * xin)
            denom = PiOverXmax * xin
            factor = np.divide(num, denom, where=denom != 0)

        yout = np.zeros_like(xout)
        eout = np.zeros_like(xout)
        for i, x in enumerate(xout):
            kernel = factor * yin * np.sin(xin * x)
            ekernel = np.square(factor * err * np.sin(xin * x))
            yout[i] = np.trapz(kernel, x=xin)
            eout[i] = np.sqrt(
                (np.diff(xin)**2 * (ekernel[1:] + ekernel[:-1]) / 2).sum())

        if kwargs.get('OmittedXrangeCorrection', False):
            self._low_x_correction(xin, yin, xout, yout, **kwargs)

        return xout, yout, eout

    # Reciprocal -> Real Space Transforms  #

    def F_to_G(self, q, fq, r, dfq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`Q[S(Q)-1]`
        to real space :math:`G_{PDFFIT}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq: uncertainty on :math:`Q[S(Q)-1]`
        :type dfq: numpy.array or list

        :return: :math:`r`, :math:`G_{PDFFIT}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        r, gr, dgr = self.fourier_transform(q, fq, r, dyin=dfq, **kwargs)
        gr *= 2. / np.pi
        dgr *= 2. / np.pi
        return r, gr, dgr

    def F_to_GK(self, q, fq, r, dfq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`Q[S(Q)-1]`
        to real space :math:`G_{Keen Version}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq: uncertainty on :math:`Q[S(Q)-1]`
        :type dfq: numpy.array or list

        :return: :math:`r`, :math:`G_{Keen Version}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        r, gr, dgr = self.F_to_G(q, fq, r, dfq, **kwargs)
        gr, dgr = self.converter.G_to_GK(r, gr, dgr, **kwargs)
        return r, gr, dgr

    def F_to_g(self, q, fq, r, dfq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`Q[S(Q)-1]`
        to real space :math:`g(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq: :math:`Q[S(Q)-1]` vector
        :type fq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq: uncertainty on :math:`Q[S(Q)-1]`
        :type dfq: numpy.array or list

        :return: :math:`r`, :math:`g(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        r, gr, dgr = self.F_to_G(q, fq, r, dfq, **kwargs)
        gr, dgr = self.converter.G_to_g(r, gr, dgr, **kwargs)
        return r, gr, dgr

    # S(Q)
    def S_to_G(self, q, sq, r, dsq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`S(Q)`
        to real space :math:`G_{PDFFIT}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dsq: :math:`S(Q)` uncertainties
        :type dsq: numpy.array or list

        :return: :math:`r`, :math:`G_{PDFFIT}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.S_to_F(q, sq, dsq)
        return self.F_to_G(q, fq, r, dfq, **kwargs)

    def S_to_GK(self, q, sq, r, dsq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`S(Q)`
        to real space :math:`G_{Keen Version}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dsq: :math:`S(Q)` uncertainties
        :type dsq: numpy.array or list

        :return: :math:`r`, :math:`G_{Keen Version}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.S_to_F(q, sq, dsq)
        return self.F_to_GK(q, fq, r, dfq, **kwargs)

    def S_to_g(self, q, sq, r, dsq=None, **kwargs):
        """
        Transforms from reciprocal space :math:`S(Q)`
        to real space :math:`g(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param sq: :math:`S(Q)` vector
        :type sq: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dsq: :math:`S(Q)` uncertainties
        :type dsq: numpy.array or list

        :return: :math:`r`, :math:`g(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.S_to_F(q, sq, dsq)
        return self.F_to_g(q, fq, r, dfq, **kwargs)

    # Keen's F(Q)
    def FK_to_G(self, q, fq_keen, r, dfq_keen=None, **kwargs):
        """
        Transforms from reciprocal space :math:`F(Q)`
        to real space :math:`G_{PDFFIT}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq_keen: :math:`F(Q)` vector
        :type fq_keen: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq_keen: :math:`F(Q)` vector uncertainties
        :type dfq_keen: numpy.array or list

        :return: :math:`r`, :math:`G_{PDFFIT}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.FK_to_F(q, fq_keen, dfq_keen, **kwargs)
        return self.F_to_G(q, fq, r, dfq, **kwargs)

    def FK_to_GK(self, q, fq_keen, r, dfq_keen=None, **kwargs):
        """
        Transforms from reciprocal space :math:`F(Q)`
        to real space :math:`G_{Keen Version}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq_keen: :math:`F(Q)` vector
        :type fq_keen: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq_keen: :math:`F(Q)` vector uncertainties
        :type dfq_keen: numpy.array or list

        :return: :math:`r`, :math:`G_{Keen Version}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.FK_to_F(q, fq_keen, dfq_keen, **kwargs)
        return self.F_to_GK(q, fq, r, dfq, **kwargs)

    def FK_to_g(self, q, fq_keen, r, dfq_keen=None, **kwargs):
        """
        Transforms from reciprocal space :math:`F(Q)`
        to real space :math:`g(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param fq_keen: :math:`F(Q)` vector
        :type fq_keen: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param dfq_keen: :math:`F(Q)` vector uncertainties
        :type dfq_keen: numpy.array or list

        :return: :math:`r`, :math:`g(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.FK_to_F(q, fq_keen, dfq_keen, **kwargs)
        return self.F_to_g(q, fq, r, dfq, **kwargs)

    # Differential cross-section = d_simga / d_Omega
    def DCS_to_G(self, q, dcs, r, ddcs=None, **kwargs):
        """
        Transforms from reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
        to real space :math:`G_{PDFFIT}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param ddcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` uncertainties
        :type ddcs: numpy.array or list

        :return: :math:`r`, :math:`G_{PDFFIT}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs, **kwargs)
        return self.F_to_G(q, fq, r, dfq, **kwargs)

    def DCS_to_GK(self, q, dcs, r, ddcs=None, **kwargs):
        """
        Transforms from reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
        to real space :math:`G_{Keen Version}(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param ddcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` uncertainties
        :type ddcs: numpy.array or list

        :return: :math:`r`, :math:`G_{Keen Version}(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs, **kwargs)
        return self.F_to_GK(q, fq, r, dfq, **kwargs)

    def DCS_to_g(self, q, dcs, r, ddcs=None, **kwargs):
        """
        Transforms from reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`
        to real space :math:`g(r)`

        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` vector
        :type dcs: numpy.array or list
        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param ddcs: :math:`\\frac{d \\sigma}{d \\Omega}(Q)` uncertainties
        :type ddcs: numpy.array or list

        :return: :math:`r`, :math:`g(r)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        fq, dfq = self.converter.DCS_to_F(q, dcs, ddcs, **kwargs)
        return self.F_to_g(q, fq, r, dfq, **kwargs)

    # Real -> Reciprocal Space Transforms  #

    # G(R) = PDF
    def G_to_F(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{PDFFIT}(r)`
        to reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{PDFFIT}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`Q[S(Q)-1]`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        return self.fourier_transform(r, gr, q, dyin=dgr, **kwargs)

    def G_to_S(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{PDFFIT}(r)`
        to reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{PDFFIT}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`S(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        q, fq, dfq = self.G_to_F(r, gr, q, dgr=dgr, **kwargs)
        sq, dsq = self.converter.F_to_S(q, fq, dfq=dfq)
        return q, sq, dsq

    def G_to_FK(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{PDFFIT}(r)`
        to reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{PDFFIT}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`F(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        q, fq, dfq = self.G_to_F(r, gr, q, dgr=dgr, **kwargs)
        fq, dfq = self.converter.F_to_FK(q, fq, dfq=dfq, **kwargs)
        return q, fq, dfq

    def G_to_DCS(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{PDFFIT}(r)`
        to reciprocal space
        :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{PDFFIT}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{PDFFIT}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        q, fq, dfq = self.G_to_F(r, gr, q, dgr=dgr, **kwargs)
        dcs, ddcs = self.converter.F_to_DCS(q, fq, dfq=dfq, **kwargs)
        return q, dcs, ddcs

    # Keen's G(r)
    def GK_to_F(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{Keen Version}(r)`
        to reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{Keen Version}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`Q[S(Q)-1]`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.GK_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_F(r, _gr, q, dgr=_dgr, **kwargs)

    def GK_to_S(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{Keen Version}(r)`
        to reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{Keen Version}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`S(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.GK_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_S(r, _gr, q, dgr=_dgr, **kwargs)

    def GK_to_FK(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{Keen Version}(r)`
        to reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{Keen Version}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`F(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.GK_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_FK(r, _gr, q, dgr=_dgr, **kwargs)

    def GK_to_DCS(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`G_{Keen Version}(r)`
        to reciprocal space :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`G_{Keen Version}(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`G_{Keen Version}(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.GK_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_DCS(r, _gr, q, dgr=_dgr, **kwargs)

    # g(r)
    def g_to_F(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`g(r)`
        to reciprocal space :math:`Q[S(Q)-1]`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`g(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`Q[S(Q)-1]`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.g_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_F(r, _gr, q, dgr=_dgr, **kwargs)

    def g_to_S(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`g(r)`
        to reciprocal space :math:`S(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`g(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`S(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.g_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_S(r, _gr, q, dgr=_dgr, **kwargs)

    def g_to_FK(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`g(r)`
        to reciprocal space :math:`F(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`g(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`F(Q)`, and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.g_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_FK(r, _gr, q, dgr=_dgr, **kwargs)

    def g_to_DCS(self, r, gr, q, dgr=None, **kwargs):
        """
        Transforms from real space :math:`g(r)`
        to reciprocal space :math:`\\frac{d \\sigma}{d \\Omega}(Q)`

        :param r: :math:`r`-space vector
        :type r: numpy.array or list
        :param gr: :math:`g(r)` vector
        :type gr: numpy.array or list
        :param q: :math:`Q`-space vector
        :type q: numpy.array or list
        :param dgr: :math:`g(r)` uncertainties
        :type dgr: numpy.array or list

        :return: :math:`Q`, :math:`\\frac{d \\sigma}{d \\Omega}(Q)`,
                 and uncertainties
        :rtype: numpy.array, numpy.array, numpy.array
        """
        _gr, _dgr = self.converter.g_to_G(r, gr, dgr=dgr, **kwargs)
        return self.G_to_DCS(r, _gr, q, dgr=_dgr, **kwargs)



import os
import setuptools
import versioneer

# Constants
THIS_DIR = os.path.dirname(__file__)

# Package description
with open("README.md", "r") as fh:
    long_description = fh.read()


# Package requirements helper
def read_requirements_from_file(filepath):
    '''Read a list of requirements from the given file and split into a
    list of strings. It is assumed that the file is a flat
    list with one requirement per line.
    :param filepath: Path to the file to read
    :return: A list of strings containing the requirements
    '''
    with open(filepath, 'rU') as req_file:
        return req_file.readlines()


setup_args = dict(
    install_requires=read_requirements_from_file(
        os.path.join(
            THIS_DIR,
            'requirements.txt')),
    tests_require=read_requirements_from_file(
        os.path.join(
            THIS_DIR,
            'requirements-dev.txt')))

authors = [
    'Marshall McDonnell (marshallmcdonnell)',
    'Mathieu Doucet (mdoucet)',
    'Ross Whitfield (rosswhitfield)',
    'Pete Peterson (peterfpeterson)',
    'Yuanpeng Zhang (Kvieta1990)',
]

setuptools.setup(
    name="pystog",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author=",".join(authors),
    author_email="mcdonnellmt@ornl.gov",
    url="https://github.com/neutrons/pystog",
    description="Manipulate total scattering functions",
    long_description_content_type="text/markdown",
    license="GPL License (version 3)",
    packages=setuptools.find_packages(exclude=["fortran"]),
    package_data={'': ['*.dat', '*.gr']},
    setup_requires=[],
    install_requires=setup_args['install_requires'],
    tests_require=setup_args['install_requires'] + setup_args['tests_require'],
    test_suite='tests',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            "pystog_cli = pystog.cli:pystog_cli",
        ]
    }
)



from pystog.stog import StoG
from pystog.converter import Converter
from pystog.transformer import Transformer
from pystog.fourier_filter import FourierFilter

__all__ = ['StoG', 'Converter', 'Transformer', 'FourierFilter', ]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



import json
import argparse
from pystog.utils import RealSpaceChoices, ReciprocalSpaceChoices


def get_cli_parser():
    """
    Create the argument parser for the CLI

    :return: Argument parser for PyStoG CLI
    :rtype: argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--json",
        type=str,
        default=None,
        help="Read JSON input file for arguments")

    parser.add_argument(
        "--density",
        type=float,
        help="Number density (atoms/angstroms^3")

    parser.add_argument(
        "-f",
        "--filename",
        nargs='*',
        action='append',
        default=list(),
        dest='filenames',
        help="Filename, qmin, qmax, yoffset, yscale, Qoffset, function type."
        "Function Types are: %s" %
        json.dumps(ReciprocalSpaceChoices))

    parser.add_argument(
        "--stem-name",
        type=str,
        dest="stem_name",
        default="merged",
        help="Stem name for output files")

    parser.add_argument(
        "--real-space-function",
        type=str,
        default="g(r)",
        dest="real_space_function",
        help="Real-space function typej. Choices are: %s" %
        json.dumps(RealSpaceChoices))

    parser.add_argument(
        "--Rmax",
        type=float,
        default=50.0,
        help="Maximum value in angstroms for real-space functions")

    parser.add_argument(
        "--Rpoints",
        type=int,
        default=5000,
        help="Number of points in R for real-space functions")

    parser.add_argument(
        "--Rdelta",
        type=float,
        default=None,
        help="Bin width to use in R for real-space functions")

    parser.add_argument(
        "--fourier-filter-cutoff",
        type=float,
        default=None,
        dest="fourier_filter_cutoff",
        help="Bin width to use in R for real-space functions")

    parser.add_argument(
        "--lorch-flag",
        action="store_true",
        default=False,
        dest="lorch_flag",
        help="Apply Lorch function")

    parser.add_argument(
        "--bcoh_sqrd",
        type=float,
        default=1.0,
        dest="bcoh_sqrd",
        help="The (sum c*bbar)^2 term needed for F(Q) and G(r) for RMC output")

    parser.add_argument(
        "--btot_sqrd",
        type=float,
        default=1.0,
        dest="btot_sqrd",
        help="The (sum c*b^2) term needed for DCS(Q) input")

    parser.add_argument(
        "--merging",
        nargs=2,
        type=float,
        default=[0.0, 1.0],
        help="Offset and Scale to apply to the merged S(Q)")

    parser.add_argument(
        "--low-q-correction",
        action='store_true',
        help="Apply low-Q correction during FT")

    return parser


def parse_cli_args(args):
    """
    Parse the CLI arguments and return a key-word dictionary with options to
    pass to StoG class

    :param args: Namespace returned from parsing the CLI arguments
    :type args: argparse.Namespace
    :return: Dictionary with options to pass to StoG class
    :rtype: dict
    """
    if not isinstance(args, argparse.Namespace):
        msg = "parse_cli_args takes a argparse.Namespace, "
        msg += "yet was given argument of type: {}"
        raise TypeError(msg.format(type(args)))

    # Get each file's info and story in dictionary
    files_info = list()
    for f in args.filenames:
        file_info = dict()
        file_info["Filename"] = f[0]
        file_info["Qmin"] = float(f[1])
        file_info["Qmax"] = float(f[2])
        file_info["Y"] = {"Offset": float(f[3]), "Scale": float(f[4])}
        file_info["X"] = {"Offset": float(f[5])}
        file_info["ReciprocalFunction"] = f[6]
        files_info.append(file_info)

    # Get general StoG options
    kwargs = {
        "Files": files_info,
        "NumberDensity": args.density,
        "Rmax": args.Rmax,
        "Rpoints": args.Rpoints,
        "FourierFilter": {
            "Cutoff": args.fourier_filter_cutoff},
        "LorchFlag": args.lorch_flag,
        "Outputs": {
            "StemName": args.stem_name},
        "Merging": {
            "Y": {
                "Offset": args.merging[0],
                "Scale": args.merging[1]}},
        "<b_coh>^2": args.bcoh_sqrd,
        "<b_tot^2>": args.btot_sqrd,
        "RealSpaceFunction": args.real_space_function,
        "OmittedXrangeCorrection": args.low_q_correction}
    if args.Rdelta:
        kwargs["Rdelta"] = args.Rdelta

    return kwargs
